#    Copyright (C) 2010 Canonical Ltd
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

"""Tests for the merge_changelog code."""

from __future__ import absolute_import

import logging
import sys

from testtools.content_type import ContentType
from testtools.content import Content

from .... import (
    merge,
    tests,
    )
from ... import debian
from .. import merge_changelog
from ....tests.features import ExecutableFeature


dpkg_mergechangelogs_feature = ExecutableFeature('dpkg-mergechangelogs')


v_111_2 = b"""\
pseudo-prog (1.1.1-2) unstable; urgency=low

  * New upstream release.
  * Awesome bug fixes.

 -- Joe Foo <joe@example.com>  Thu, 28 Jan 2010 10:45:44 +0000

""".splitlines(True)


v_111_2b = b"""\
pseudo-prog (1.1.1-2) unstable; urgency=low

  * New upstream release.
  * Awesome bug fixes.
  * But more is better

 -- Joe Foo <joe@example.com>  Thu, 28 Jan 2010 10:45:44 +0000

""".splitlines(True)


v_111_2c = b"""\
pseudo-prog (1.1.1-2) unstable; urgency=low

  * New upstream release.
  * Yet another content for 1.1.1-2

 -- Joe Foo <joe@example.com>  Thu, 28 Jan 2010 10:45:44 +0000

""".splitlines(True)


# Merge of 2b and 2c using 2 as the base (b adds a line, c adds a line and
# deletes a line).
v_111_2bc = b"""\
pseudo-prog (1.1.1-2) unstable; urgency=low

  * New upstream release.
  * Yet another content for 1.1.1-2
  * But more is better

 -- Joe Foo <joe@example.com>  Thu, 28 Jan 2010 10:45:44 +0000

""".splitlines(True)


# Merge of 2b and 2c using an empty base. (As calculated by
# dpkg-mergechangelogs.)
v_111_2bc_empty_base = b"""\
pseudo-prog (1.1.1-2) unstable; urgency=low

  * New upstream release.
<<<<<<<
  * Awesome bug fixes.
=======
  * Yet another content for 1.1.1-2
>>>>>>>
  * But more is better

 -- Joe Foo <joe@example.com>  Thu, 28 Jan 2010 10:45:44 +0000

""".splitlines(True)


v_112_1 = b"""\
pseudo-prog (1.1.2-1) unstable; urgency=low

  * New upstream release.
  * No bug fixes :(

 -- Barry Foo <barry@example.com>  Thu, 27 Jan 2010 10:45:44 +0000

""".splitlines(True)


v_001_1 = b"""\
pseudo-prog (0.0.1-1) unstable; urgency=low

  * New project released!!!!
  * No bugs evar

 -- Barry Foo <barry@example.com>  Thu, 27 Jan 2010 10:00:44 +0000

""".splitlines(True)


# Backports from current testtools so that we remain compatible with testtools
# 0.9.2 (the version in lucid).
UTF8_TEXT = ContentType('text', 'plain', {'charset': 'utf8'})


class TestMergeChangelog(tests.TestCase):

    _test_needs_features = [dpkg_mergechangelogs_feature]

    def setUp(self):
        super(tests.TestCase, self).setUp()
        # Intercept warnings from merge_changelog's logger: this is where 
        self.logged_warnings = self.make_utf8_encoded_stringio()
        self.addCleanup(self.addMergeChangelogWarningsDetail)
        handler = logging.StreamHandler(self.logged_warnings)
        handler.setLevel(logging.WARNING)
        logger = logging.getLogger('breezy.plugins.debian.merge_changelog')
        logger.addHandler(handler)
        self.addCleanup(logger.removeHandler, handler)
        self.overrideAttr(logger, 'propagate', False)

    def addMergeChangelogWarningsDetail(self):
        warnings_log = self.logged_warnings.getvalue()
        if warnings_log:
            self.addDetail(
                'merge_changelog warnings',
                Content(UTF8_TEXT, lambda: [warnings_log]))

    def assertMergeChangelog(self, expected_lines, this_lines, other_lines,
                             base_lines=[], conflicted=False,
                             possible_error=False):
        status, merged_lines = merge_changelog.merge_changelog(
                                    this_lines, other_lines, base_lines)
        if possible_error and status == "not_applicable":
            self.assertContainsRe(
                self.logged_warnings.getvalue(),
                "(?m)dpkg-mergechangelogs failed with status \\d+$")
            return False
        if conflicted:
            self.assertEqual('conflicted', status)
        else:
            self.assertEqual('success', status)
        self.assertEqualDiff(b''.join(expected_lines), b''.join(merged_lines))
        return True

    def test_merge_by_version(self):
        this_lines = v_111_2 + v_001_1
        other_lines = v_112_1 + v_001_1
        expected_lines = v_112_1 + v_111_2 + v_001_1
        self.assertMergeChangelog(expected_lines, this_lines, other_lines)
        self.assertMergeChangelog(expected_lines, other_lines, this_lines)

    def test_this_shorter(self):
        self.assertMergeChangelog(
            v_112_1 + v_111_2 + v_001_1,
            this_lines=v_111_2,
            other_lines=v_112_1 + v_001_1,
            base_lines=[])
        self.assertMergeChangelog(
            v_112_1 + v_111_2 + v_001_1,
            this_lines=v_001_1,
            other_lines=v_112_1 + v_111_2,
            base_lines=[])

    def test_other_shorter(self):
        self.assertMergeChangelog(
            v_112_1 + v_111_2 + v_001_1,
            this_lines=v_112_1 + v_001_1,
            other_lines=v_111_2,
            base_lines=[])
        self.assertMergeChangelog(
            v_112_1 + v_111_2 + v_001_1,
            this_lines=v_112_1 + v_111_2,
            other_lines=v_001_1,
            base_lines=[])

    def test_unsorted(self):
        # The order of entries being merged is unchanged, even if they are not
        # properly sorted.  (This is a merge tool, not a reformatting tool.)
        self.assertMergeChangelog(v_111_2 + v_001_1,
                                  this_lines=v_111_2 + v_001_1,
                                  other_lines=[],
                                  base_lines=[])

    def test_3way_merge(self):
        # Check that if one of THIS or OTHER matches BASE, then we select the
        # other content
        self.assertMergeChangelog(expected_lines=v_111_2,
                                  this_lines=v_111_2, other_lines=v_111_2b,
                                  base_lines=v_111_2b)
        self.assertMergeChangelog(expected_lines=v_111_2b,
                                  this_lines=v_111_2, other_lines=v_111_2b,
                                  base_lines=v_111_2)

    def test_3way_conflicted(self):
        self.assertMergeChangelog(
            expected_lines=v_111_2bc,
            this_lines=v_111_2b, other_lines=v_111_2c,
            base_lines=v_111_2)
        self.assertMergeChangelog(
            expected_lines=v_111_2bc_empty_base,
            this_lines=v_111_2b, other_lines=v_111_2c,
            base_lines=[],
            conflicted=True)

    def test_not_valid_changelog(self):
        invalid_changelog = b"""\
pseudo-prog (1.1.1-2) unstable; urgency=low

  * New upstream release.
  * Awesome bug fixes.

 -- Thu, 28 Jan 2010 10:45:44 +0000

""".splitlines(True)
        # invalid_changelog is missing the author, but dpkg-mergechangelogs
        # copes gracefully with invalid input.
        status, lines = merge_changelog.merge_changelog(
            invalid_changelog, v_111_2, v_111_2)
        self.assertEqual('success', status)
        # XXX: ideally we'd expect ''.join(lines) ==
        # ''.join(invalid_changelog), but dpkg-mergechangelogs appears to lose
        # the final line in these examples.
        # <https://bugs.launchpad.net/ubuntu/+source/dpkg/+bug/815704>
        #  - Andrew Bennetts, 25 July 2011.
        # self.assertEqual(''.join(invalid_changelog), ''.join(lines))
        self.assertMergeChangelog(v_112_1 +
                                  [b'<<<<<<<\n'] +
                                  v_111_2 +
                                  [b'=======\n>>>>>>>\n'],
                                  this_lines=v_111_2,
                                  other_lines=v_112_1,
                                  base_lines=invalid_changelog,
                                  conflicted=True
                                  )

    def test_invalid_version_starting_non_digit(self):
        """Invalid version without digit first is rejected or correctly merged

        Versions of dpkg prior to 1.16.0.1 merge such changelogs correctly,
        however then a stricter check was introduced that aborts the script.
        In that case, the result should not be a success with a zero byte
        merge result file. See lp:893495 for such an issue.
        """
        invalid_changelog = b"""\
pseudo-prog (ss-0) unstable; urgency=low

  * New project released!!!!

 -- Barry Foo <barry@example.com>  Thu, 28 Jan 2010 10:00:44 +0000

""".splitlines(True)
        handled = self.assertMergeChangelog(
            expected_lines=v_112_1 + v_111_2 + invalid_changelog,
            this_lines=v_112_1 + invalid_changelog,
            other_lines=v_111_2 + invalid_changelog,
            base_lines=invalid_changelog,
            possible_error=True)
        if not handled:
            # Can't assert on the exact message as it depends on the locale
            self.assertContainsRe(
                self.logged_warnings.getvalue(),
                "dpkg-mergechangelogs: .*ss-0( is not a valid version)?")

    def test_invalid_version_non_ascii(self):
        """Invalid version with non-ascii data is rejected or correctly merged

        Such a version has always been treated as invalid so fails
        consistently across dpkg versions currently.
        """
        invalid_changelog = b"""\
pseudo-prog (\xc2\xa7) unstable; urgency=low

  * New project released!!!!

 -- Barry Foo <barry@example.com>  Thu, 28 Jan 2010 10:00:44 +0000

""".splitlines(True)
        handled = self.assertMergeChangelog(
            expected_lines=v_112_1 + v_111_2 + invalid_changelog,
            this_lines=v_112_1 + invalid_changelog,
            other_lines=v_111_2 + invalid_changelog,
            base_lines=invalid_changelog,
            possible_error=True)
        if not handled:
            # Can't assert on the exact message as it depends on the locale
            self.assertContainsRe(
                self.logged_warnings.getvalue(),
                "dpkg-mergechangelogs: .*( is not a valid version)?")


class TestChangelogHook(tests.TestCaseWithMemoryTransport):

    _test_needs_features = [dpkg_mergechangelogs_feature]

    def make_params(self):
        builder = self.make_branch_builder('source')
        builder.start_series()
        builder.build_snapshot(None, [
            ('add', ('', b'TREE_ROOT', 'directory', None)),
            ('add', ('debian', b'deb-id', 'directory', None)),
            ('add', ('debian/changelog', b'c-id', 'file', b'')),
            ('add', ('changelog', b'o-id', 'file', b'')),
            ])
        builder.finish_series()
        the_branch = builder.get_branch()

        tree = the_branch.create_memorytree()
        tree.lock_write()
        self.addCleanup(tree.unlock)

        class FakeMerger(object):

            def __init__(self, this_tree):
                self.this_tree = this_tree

            def get_lines(self, tree, path):
                return tree.get_file_lines(path)

        merger = FakeMerger(tree)
        params_cls = merge.MergeFileHookParams
        from inspect import signature
        params_cls_arg_count = len(signature(params_cls).parameters) + 1
        # Older versions of Breezy required a file_id to be specified.
        if params_cls_arg_count == 7:
            params = params_cls(
                merger,
                ('debian/changelog', 'debian/changelog', 'debian/changelog'),
                None, 'file', 'file', 'this')
        elif params_cls_arg_count == 8:
            params = params_cls(
                merger, b'c-id',
                ('debian/changelog', 'debian/changelog', 'debian/changelog'),
                None, 'file', 'file', 'this')
        else:
            raise AssertionError(params_cls_arg_count)

        return params, merger

    def test_changelog_merge_hook_successful(self):
        params, merger = self.make_params()
        params.other_lines = [b'']
        params.base_lines = [b'']
        file_merger = debian.changelog_merge_hook_factory(merger)
        result, new_content = file_merger.merge_text(params)
        self.assertEqual('success', result)
        # We ignore the new_content, as we test that elsewhere



#    test_source_distiller.py -- Getting the source to build from a branch
#    Copyright (C) 2008 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from __future__ import absolute_import

import os

from debian.changelog import Version

from ....errors import (
    FileExists,
    )

from ..upstream import MissingUpstreamTarball
from ..source_distiller import (
    FullSourceDistiller,
    MergeModeDistiller,
    NativeSourceDistiller,
    )
from . import (
    SourcePackageBuilder,
    TestCaseWithTransport,
    )
from .test_upstream import (
    _MissingUpstreamProvider,
    _SimpleUpstreamProvider,
    _TouchUpstreamProvider,
    )


class NativeSourceDistillerTests(TestCaseWithTransport):

    def test_distill_target_exists(self):
        wt = self.make_branch_and_tree(".")
        wt.lock_read()
        self.addCleanup(wt.unlock)
        sd = NativeSourceDistiller(wt, '')
        self.build_tree(['target/'])
        self.assertRaises(FileExists, sd.distill, 'target')

    def test_distill_revision_tree(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(['a'])
        wt.lock_write()
        self.addCleanup(wt.unlock)
        wt.add(['a'])
        wt.commit("one")
        rev_tree = wt.basis_tree()
        sd = NativeSourceDistiller(rev_tree, '')
        sd.distill('target')
        self.assertPathExists('target')
        self.assertPathExists('target/a')

    def test_distill_removes_builddeb_dir(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(
            ['a', '.bzr-builddeb/', '.bzr-builddeb/default.conf'])
        wt.lock_write()
        self.addCleanup(wt.unlock)
        wt.add(['a', '.bzr-builddeb/',
                '.bzr-builddeb/default.conf'])
        wt.commit("one")
        rev_tree = wt.basis_tree()
        sd = NativeSourceDistiller(rev_tree, '')
        sd.distill('target')
        self.assertPathExists('target')
        self.assertPathExists('target/a')
        self.assertPathDoesNotExist('target/.bzr-builddeb')

    def test_distill_working_tree_with_symlinks(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(['a'])
        os.symlink('a', 'b')
        wt.lock_write()
        self.addCleanup(wt.unlock)
        wt.add(['a', 'b'])
        sd = NativeSourceDistiller(wt, '')
        sd.distill('target')
        self.assertPathExists('target')
        self.assertPathExists('target/a')
        self.assertPathExists('target/b')


class FullSourceDistillerTests(TestCaseWithTransport):

    def test_distill_target_exists(self):
        wt = self.make_branch_and_tree(".")
        wt.lock_read()
        self.addCleanup(wt.unlock)
        sd = FullSourceDistiller(wt, '', None)
        self.build_tree(['target/'])
        self.assertRaises(FileExists, sd.distill, 'target')

    def test_distill_no_tarball(self):
        wt = self.make_branch_and_tree(".")
        wt.lock_read()
        self.addCleanup(wt.unlock)
        sd = FullSourceDistiller(wt, '', _MissingUpstreamProvider())
        self.assertRaises(MissingUpstreamTarball, sd.distill, 'target')

    def test_distill_tarball_exists(self):
        wt = self.make_branch_and_tree(".")
        wt.lock_read()
        self.addCleanup(wt.unlock)
        sd = FullSourceDistiller(wt, '', _TouchUpstreamProvider('tarball'))
        sd.distill('target')
        self.assertPathExists('tarball')

    def test_distill_revision_tree(self):
        wt = self.make_branch_and_tree(".")
        wt.lock_write()
        self.addCleanup(wt.unlock)
        self.build_tree(['a', '.bzr-builddeb'])
        wt.add(['a', '.bzr-builddeb'])
        sd = FullSourceDistiller(wt, '', _TouchUpstreamProvider('tarball'))
        sd.distill('target')
        self.assertPathExists('tarball')
        self.assertPathExists('target')
        self.assertPathExists('target/a')
        self.assertPathDoesNotExist('target/.bzr-builddeb')

    def test_distill_working_tree_with_symlinks(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(['a'])
        os.symlink('a', 'b')
        wt.lock_write()
        self.addCleanup(wt.unlock)
        wt.add(['a', 'b'])
        sd = FullSourceDistiller(wt, '', _TouchUpstreamProvider('tarball'))
        sd.distill('target')
        self.assertPathExists('target')
        self.assertPathExists('target/a')
        self.assertPathExists('target/b')


class MergeModeDistillerTests(TestCaseWithTransport):

    def make_tarball(self, name, version):
        builder = SourcePackageBuilder(name, version)
        builder.add_upstream_file("a")
        builder.add_default_control()
        builder.build()

    def test_distill_target_exists(self):
        wt = self.make_branch_and_tree(".")
        wt.lock_read()
        self.addCleanup(wt.unlock)
        sd = MergeModeDistiller(wt, '', None)
        self.build_tree(['target/'])
        self.assertRaises(FileExists, sd.distill, 'target')

    def test_distill_no_tarball(self):
        wt = self.make_branch_and_tree(".")
        wt.lock_read()
        self.addCleanup(wt.unlock)
        sd = MergeModeDistiller(
            wt, '', _SimpleUpstreamProvider("package", "0.1-1", "tarballs"))
        self.assertRaises(MissingUpstreamTarball, sd.distill, 'target')

    def test_distill_tarball_exists(self):
        wt = self.make_branch_and_tree(".")
        wt.lock_read()
        self.addCleanup(wt.unlock)
        name = "package"
        version = Version("0.1-1")
        self.make_tarball(name, version)
        sd = MergeModeDistiller(
            wt, '', _SimpleUpstreamProvider(name, version.upstream_version, "."))
        sd.distill('target/foo')
        self.assertPathExists(
            'target/%s_%s.orig.tar.gz' % (name, version.upstream_version))
        self.assertPathExists('target/foo/a')

    def test_distill_exports_branch(self):
        wt = self.make_branch_and_tree('.')
        wt.lock_write()
        self.addCleanup(wt.unlock)
        self.build_tree(['debian/', 'debian/a', '.bzr-builddeb'])
        wt.add(['debian/', 'debian/a', '.bzr-builddeb'])
        name = "package"
        version = Version("0.1-1")
        self.make_tarball(name, version)
        sd = MergeModeDistiller(
            wt, '', _SimpleUpstreamProvider(name, version.upstream_version, "."))
        sd.distill('target/')
        self.assertPathExists('target/debian/a')
        self.assertPathDoesNotExist('target/.bzr-builddeb')

    def test_distill_removes_debian(self):
        wt = self.make_branch_and_tree('.')
        wt.lock_write()
        self.addCleanup(wt.unlock)
        self.build_tree(['debian/', 'debian/a', '.bzr-builddeb'])
        wt.add(['debian/', 'debian/a', '.bzr-builddeb'])
        name = "package"
        version = Version("0.1-1")
        builder = SourcePackageBuilder(name, version)
        builder.add_upstream_file("a")
        builder.add_upstream_file("debian/foo")
        builder.add_default_control()
        builder.build()
        sd = MergeModeDistiller(
            wt, '', _SimpleUpstreamProvider(name, version.upstream_version, "."))
        sd.distill('target/')
        self.assertPathExists('target/a')
        self.assertPathDoesNotExist('target/debian/foo')

    def test_distill_top_level(self):
        wt = self.make_branch_and_tree('.')
        wt.lock_write()
        self.addCleanup(wt.unlock)
        self.build_tree(['b', '.bzr-builddeb'])
        wt.add(['b', '.bzr-builddeb'])
        name = "package"
        version = Version("0.1-1")
        self.make_tarball(name, version)
        sd = MergeModeDistiller(
            wt, '', _SimpleUpstreamProvider(name, version.upstream_version, "."),
            top_level=True)
        sd.distill('target/')
        self.assertPathExists('target/a')
        self.assertPathExists('target/debian/b')
        self.assertPathDoesNotExist('target/debian/.bzr-builddeb')
        self.assertPathDoesNotExist('target/.bzr-builddeb')
        self.assertPathDoesNotExist('target/b')

    def test_distill_use_existing(self):
        wt = self.make_branch_and_tree('.')
        wt.lock_write()
        self.addCleanup(wt.unlock)
        self.build_tree(['debian/', 'debian/a', '.bzr-builddeb'])
        wt.add(['debian/', 'debian/a', '.bzr-builddeb'])
        name = "package"
        version = Version("0.1-1")
        self.make_tarball(name, version)
        sd = MergeModeDistiller(
            wt, '', _SimpleUpstreamProvider(name, version.upstream_version, "."),
            use_existing=True)
        self.build_tree(
            ['target/', 'target/b', 'target/debian/', 'target/debian/b'])
        sd.distill('target/')
        self.assertPathExists('target/b')
        self.assertPathExists('target/debian/a')
        self.assertPathDoesNotExist('target/a')
        self.assertPathDoesNotExist('target/debian/b')

    def test_distill_working_tree_with_symlinks(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(['debian/', 'debian/a'])
        os.symlink('a', 'debian/b')
        wt.lock_write()
        self.addCleanup(wt.unlock)
        wt.add(['debian', 'debian/a', 'debian/b'])
        name = "package"
        version = Version("0.1-1")
        self.make_tarball(name, version)
        sd = MergeModeDistiller(
            wt, '', _SimpleUpstreamProvider(name, version.upstream_version, "."))
        sd.distill('target')
        self.assertPathExists('target')
        self.assertPathExists('target/debian/a')
        self.assertPathExists('target/debian/b')



#    __init__.py -- The plugin for bzr
#    Copyright (C) 2009-2018 Jelmer Vernooij <jelmer@jelmer.uk
#
#    This file is part of brz-debian.
#
#    brz-debian is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    brz-debian is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with brz-debian; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import posixpath
from typing import List, Optional

from breezy.errors import BzrError
from breezy.tree import Tree
from debian.changelog import Changelog
from debmutate.changelog import new_changelog_entries, strip_changelog_message

from . import tree_debian_tag_name


class UnreleasedChanges(BzrError):

    _fmt = "%(path)s says it's UNRELEASED."

    def __init__(self, path):
        BzrError.__init__(self)
        self.path = path


def changelog_changes(
        tree: Tree, basis_tree: Tree,
        cl_path: str = 'debian/changelog') -> List[str]:
    changes = []
    for change in tree.iter_changes(
            basis_tree, specific_files=[cl_path]):
        paths = change.path
        changed_content = change.changed_content
        versioned = change.versioned
        kind = change.kind
        # Content not changed
        if not changed_content:
            return None
        # Not versioned in new tree
        if not versioned[1]:
            return None
        # Not a file in one tree
        if kind[0] != 'file' or kind[1] != 'file':
            return None

        old_text = basis_tree.get_file_lines(paths[0])
        new_text = tree.get_file_lines(paths[1])
        changes.extend(new_changelog_entries(old_text, new_text))
    return changes


def changelog_commit_message(
        tree: Tree, basis_tree: Tree,
        path: str = 'debian/changelog') -> Optional[str]:
    changes = changelog_changes(tree, basis_tree, path)
    if not changes:
        return None

    return ''.join(strip_changelog_message(changes))


def debcommit(tree, committer=None, subpath="", paths=None):
    message = changelog_commit_message(
        tree, tree.basis_tree(),
        path=posixpath.join(subpath, "debian/changelog")
    )
    if paths:
        specific_files = [posixpath.join(subpath, p) for p in paths]
    elif subpath:
        specific_files = [subpath]
    else:
        specific_files = None
    tree.commit(
        committer=committer, message=message, specific_files=specific_files)


def debcommit_release(tree, committer=None, subpath="", message=None):
    tag_name = tree_debian_tag_name(tree, tree.branch, subpath=subpath)
    cl_path = posixpath.join(subpath, "debian/changelog")
    if tag_name is None:
        raise UnreleasedChanges(cl_path)
    if message is None:
        cl = Changelog(tree.get_file(cl_path), max_blocks=1)
        message = "releasing package %s version %s" % (
                cl[0].package, cl[0].version)
    revid = tree.commit(committer=committer, message=message)
    tree.branch.tags.set_tag(tag_name, revid)
    return tag_name



from __future__ import absolute_import

import os
import sys
import subprocess

from ... import (
    controldir,
    errors as bzr_errors,
    revision as mod_revision,
    trace,
    transport,
    workingtree,
    )

from . import (
    default_orig_dir,
    import_dsc,
    upstream,
    util,
    )


def _get_tree(package_name):
    try:
        tree = workingtree.WorkingTree.open(".")
    except bzr_errors.NotBranchError:
        if os.path.exists(package_name):
            raise bzr_errors.BzrCommandError(
                "Either run the command from an "
                "existing branch of upstream, or move %s aside "
                "and a new branch will be created there."
                % package_name)
        to_transport = transport.get_transport(package_name)
        tree = to_transport.ensure_base()
        try:
            a_controldir = controldir.ControlDir.open_from_transport(
                to_transport)
        except bzr_errors.NotBranchError:
            # really a NotBranchError...
            create_branch = controldir.ControlDir.create_branch_convenience
            branch = create_branch(to_transport.base,
                                   possible_transports=[to_transport])
            a_controldir = branch.controldir
        else:
            if a_controldir.has_branch():
                raise bzr_errors.AlreadyBranchError(package_name)
            branch = a_controldir.create_branch()
            a_controldir.create_workingtree()
        try:
            tree = a_controldir.open_workingtree()
        except bzr_errors.NoWorkingTree:
            tree = a_controldir.create_workingtree()
    return tree


def _get_tarballs(tree, subpath, tarball, package_name, version):
    from .repack_tarball import repack_tarball
    config = util.debuild_config(tree, subpath)
    orig_dir = config.orig_dir or default_orig_dir
    orig_dir = os.path.join(tree.basedir, orig_dir)
    if not os.path.exists(orig_dir):
        os.makedirs(orig_dir)
    format = None
    if tarball.endswith(".tar.bz2") or tarball.endswith(".tbz2"):
        format = "bz2"
    elif tarball.endswith(".tar.xz"):
        format = "xz"
    dest_name = util.tarball_name(package_name, version, None, format=format)
    trace.note("Fetching tarball")
    repack_tarball(tarball, dest_name, target_dir=orig_dir)
    provider = upstream.UpstreamProvider(
        package_name, version, orig_dir, [])
    orig_files = provider.provide(os.path.join(tree.basedir, ".."))
    ret = []
    for filename, component in orig_files:
        ret.append((filename, component, util.md5sum_filename(filename)))
    return ret


def import_upstream(tarball, package_name, version, use_pristine_tar=True):
    tree = _get_tree(package_name)
    if tree.branch.last_revision() != mod_revision.NULL_REVISION:
        parents = {None: [tree.branch.last_revision()]}
    else:
        parents = {}
    tarball_filenames = _get_tarballs(
        tree, '', tarball, package_name, version)
    db = import_dsc.DistributionBranch(
        tree.branch, tree.branch, tree=tree, pristine_upstream_tree=tree)
    dbs = import_dsc.DistributionBranchSet()
    dbs.add_branch(db)
    db.import_upstream_tarballs(
        tarball_filenames, package_name, version, parents,
        force_pristine_tar=use_pristine_tar)
    return tree



#    import_dsc.py -- Import a series of .dsc files.
#    Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
#              (C) 2008 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

from debian.changelog import Version

from contextlib import ExitStack
import os
import subprocess
import tempfile

from ... import osutils
from ...trace import mutter

from .util import (
    FORMAT_1_0,
    FORMAT_3_0_QUILT,
    FORMAT_3_0_NATIVE,
    component_from_orig_tarball,
    subprocess_setup,
)


class SourceExtractor(object):
    """A class to extract a source package to its constituent parts"""

    def __init__(self, dsc_path, dsc):
        self.dsc_path = dsc_path
        self.dsc = dsc
        self.extracted_upstream = None
        self.extracted_debianised = None
        self.unextracted_debian_md5 = None
        self.upstream_tarballs = []
        self.exit_stack = ExitStack()

    def extract(self):
        """Extract the package to a new temporary directory."""
        raise NotImplementedError(self.extract)

    def __enter__(self):
        self.exit_stack.__enter__()
        self.extract()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.exit_stack.__exit__(exc_type, exc_val, exc_tb)
        return False


class OneZeroSourceExtractor(SourceExtractor):
    """Source extract for the "1.0" source format."""

    def extract(self):
        """Extract the package to a new temporary directory."""
        tempdir = self.exit_stack.enter_context(tempfile.TemporaryDirectory())
        dsc_filename = os.path.abspath(self.dsc_path)
        proc = subprocess.Popen(
            "dpkg-source -su -x %s" % (dsc_filename,), shell=True,
            cwd=tempdir, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT, preexec_fn=subprocess_setup)
        (stdout, _) = proc.communicate()
        assert proc.returncode == 0, "dpkg-source -x failed, output:\n%s" % \
            (stdout,)
        name = self.dsc['Source']
        version = Version(self.dsc['Version'])
        self.extracted_upstream = os.path.join(
            tempdir, "%s-%s.orig" % (name, str(version.upstream_version)))
        self.extracted_debianised = os.path.join(
            tempdir, "%s-%s" % (name, str(version.upstream_version)))
        if not os.path.exists(self.extracted_upstream):
            mutter("It's a native package")
            self.extracted_upstream = None
        for part in self.dsc['files']:
            if self.extracted_upstream is None:
                if part['name'].endswith(".tar.gz"):
                    self.unextracted_debian_md5 = part['md5sum']
            else:
                if part['name'].endswith(".orig.tar.gz"):
                    self.upstream_tarballs.append((os.path.abspath(
                            os.path.join(
                                osutils.dirname(self.dsc_path),
                                part['name'])),
                            component_from_orig_tarball(
                                part['name'], name,
                                str(version.upstream_version)),
                            str(part['md5sum'])))
                elif part['name'].endswith(".diff.gz"):
                    self.unextracted_debian_md5 = part['md5sum']


class ThreeDotZeroNativeSourceExtractor(SourceExtractor):
    """Source extractor for the "3.0 (native)" source format."""

    def extract(self):
        tempdir = self.exit_stack.enter_context(tempfile.TemporaryDirectory())
        dsc_filename = os.path.abspath(self.dsc_path)
        proc = subprocess.Popen(
            "dpkg-source -x %s" % (dsc_filename,), shell=True,
            cwd=tempdir, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT, preexec_fn=subprocess_setup)
        (stdout, _) = proc.communicate()
        assert proc.returncode == 0, "dpkg-source -x failed, output:\n%s" % \
            (stdout,)
        name = self.dsc['Source']
        version = Version(self.dsc['Version'])
        self.extracted_debianised = os.path.join(
            tempdir, "%s-%s" % (name, str(version.upstream_version)))
        self.extracted_upstream = None
        for part in self.dsc['files']:
            if (part['name'].endswith(".tar.gz")
                    or part['name'].endswith(".tar.bz2")
                    or part['name'].endswith(".tar.xz")):
                self.unextracted_debian_md5 = part['md5sum']


class ThreeDotZeroQuiltSourceExtractor(SourceExtractor):
    """Source extractor for the "3.0 (quilt)" source format."""

    def extract(self):
        tempdir = self.exit_stack.enter_context(tempfile.TemporaryDirectory())
        dsc_filename = os.path.abspath(self.dsc_path)
        proc = subprocess.Popen(
            "dpkg-source --skip-debianization -x %s"
            % (dsc_filename,), shell=True,
            cwd=tempdir, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT, preexec_fn=subprocess_setup)
        (stdout, _) = proc.communicate()
        assert proc.returncode == 0, "dpkg-source -x failed, output:\n%s" % \
            (stdout,)
        name = self.dsc['Source']
        version = Version(self.dsc['Version'])
        self.extracted_debianised = os.path.join(
            tempdir, "%s-%s" % (name, str(version.upstream_version)))
        self.extracted_upstream = self.extracted_debianised + ".orig"
        os.rename(self.extracted_debianised, self.extracted_upstream)
        proc = subprocess.Popen(
            "dpkg-source -x %s" % (dsc_filename,), shell=True,
            cwd=tempdir, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT, preexec_fn=subprocess_setup)
        (stdout, _) = proc.communicate()
        assert proc.returncode == 0, "dpkg-source -x failed, output:\n%s" % \
            (stdout,)
        # Check that there are no unreadable files extracted.
        subprocess.call(
            ["find", self.extracted_upstream, "-perm",
             "0000", "-exec", "chmod", "644", "{}", ";"])
        subprocess.call(
            ["find", self.extracted_debianised, "-perm",
             "0000", "-exec", "chmod", "644", "{}", ";"])
        for part in self.dsc['files']:
            if part['name'].startswith(
                    "%s_%s.orig" % (name, str(version.upstream_version))):
                self.upstream_tarballs.append((
                    os.path.abspath(os.path.join(
                        osutils.dirname(self.dsc_path), part['name'])),
                    component_from_orig_tarball(
                        part['name'], name, str(version.upstream_version)),
                    str(part['md5sum'])))
            elif (part['name'].endswith(".debian.tar.gz")
                    or part['name'].endswith(".debian.tar.bz2")
                    or part['name'].endswith(".debian.tar.xz")):
                self.unextracted_debian_md5 = str(part['md5sum'])
        assert self.upstream_tarballs is not None, \
            "Can't handle non gz|bz2|xz tarballs yet"
        assert self.unextracted_debian_md5 is not None, \
            "Can't handle non gz|bz2|xz tarballs yet"


SOURCE_EXTRACTORS = {}
SOURCE_EXTRACTORS[FORMAT_1_0] = OneZeroSourceExtractor
SOURCE_EXTRACTORS[FORMAT_3_0_NATIVE] = ThreeDotZeroNativeSourceExtractor
SOURCE_EXTRACTORS[FORMAT_3_0_QUILT] = ThreeDotZeroQuiltSourceExtractor


def extract(dsc_filename, dsc):
    format = dsc.get('Format', FORMAT_1_0).strip()
    extractor_cls = SOURCE_EXTRACTORS.get(format)
    if extractor_cls is None:
        raise AssertionError(
            "Don't know how to import source format %s yet" % format)
    return extractor_cls(dsc_filename, dsc)



#    test_util.py -- Testsuite for builddeb util.py
#    Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

import bz2
import gzip
import hashlib
import os
import shutil
import tarfile

from debian.changelog import Changelog, Version
from debmutate.changelog import strip_changelog_message

from ..config import (
    BUILD_TYPE_MERGE,
    BUILD_TYPE_NATIVE,
    BUILD_TYPE_NORMAL,
    )
from . import (
    LzmaFeature,
    SourcePackageBuilder,
    TestCaseInTempDir,
    TestCaseWithTransport,
    )
from ..util import (
    AddChangelogError,
    InconsistentSourceFormatError,
    NoPreviousUpload,
    changelog_find_previous_upload,
    component_from_orig_tarball,
    dget,
    dget_changes,
    extract_orig_tarballs,
    find_bugs_fixed,
    find_changelog,
    find_extra_authors,
    find_thanks,
    get_build_architecture,
    get_files_excluded,
    get_commit_info_from_changelog,
    guess_build_type,
    lookup_distribution,
    move_file_if_different,
    get_parent_dir,
    recursive_copy,
    safe_decode,
    suite_to_distribution,
    tarball_name,
    tree_contains_upstream_source,
    tree_get_source_format,
    write_if_different,
    MissingChangelogError,
    )

from .... import errors as bzr_errors
from six import text_type
from ....tests import (
    TestCase,
    )
from ....tests.features import (
    SymlinkFeature,
    ModuleAvailableFeature,
    )


class RecursiveCopyTests(TestCaseInTempDir):

    def test_recursive_copy(self):
        os.mkdir('a')
        os.mkdir('b')
        os.mkdir('c')
        os.mkdir('a/d')
        os.mkdir('a/d/e')
        with open('a/f', 'w') as f:
            f.write('f')
        os.mkdir('b/g')
        recursive_copy('a', 'b')
        self.assertPathExists('a')
        self.assertPathExists('b')
        self.assertPathExists('c')
        self.assertPathExists('b/d')
        self.assertPathExists('b/d/e')
        self.assertPathExists('b/f')
        self.assertPathExists('a/d')
        self.assertPathExists('a/d/e')
        self.assertPathExists('a/f')

    def test_recursive_copy_symlink(self):
        os.mkdir('a')
        os.symlink('c', 'a/link')
        os.mkdir('b')
        recursive_copy('a', 'b')
        self.assertPathExists('b')
        self.assertPathExists('b/link')
        self.assertEqual('c', os.readlink('b/link'))


class SafeDecodeTests(TestCase):

    def assertSafeDecode(self, expected, val):
        self.assertEqual(expected, safe_decode(val))

    def test_utf8(self):
        self.assertSafeDecode(u'ascii', 'ascii')
        self.assertSafeDecode(u'\xe7', b'\xc3\xa7')

    def test_iso_8859_1(self):
        self.assertSafeDecode(u'\xe7', b'\xe7')


cl_block1 = """\
bzr-builddeb (0.17) unstable; urgency=low

  [ James Westby ]
  * Pass max_blocks=1 when constructing changelogs as that is all that is
    needed currently.

 -- James Westby <jw+debian@jameswestby.net>  Sun, 17 Jun 2007 18:48:28 +0100

"""


class FindChangelogTests(TestCaseWithTransport):

    def write_changelog(self, filename):
        f = open(filename, 'w')
        try:
            f.write(cl_block1)
            f.write("""\
bzr-builddeb (0.16.2) unstable; urgency=low

  * loosen the dependency on bzr. bzr-builddeb seems to be not be broken
    by bzr version 0.17, so remove the upper bound of the dependency.

 -- Reinhard Tartler <siretart@tauware.de>  Tue, 12 Jun 2007 19:45:38 +0100
""")
        finally:
            f.close()

    def test_find_changelog_std(self):
        tree = self.make_branch_and_tree('.')
        os.mkdir('debian')
        self.write_changelog('debian/changelog')
        tree.add(['debian', 'debian/changelog'])
        (cl, lq) = find_changelog(tree, merge=False)
        self.assertEqual(str(cl), cl_block1)
        self.assertEqual(lq, False)

    def test_find_changelog_merge(self):
        tree = self.make_branch_and_tree('.')
        os.mkdir('debian')
        self.write_changelog('debian/changelog')
        tree.add(['debian', 'debian/changelog'])
        (cl, lq) = find_changelog(tree, merge=True)
        self.assertEqual(str(cl), cl_block1)
        self.assertEqual(lq, False)

    def test_find_changelog_merge_lq(self):
        tree = self.make_branch_and_tree('.')
        self.write_changelog('changelog')
        tree.add(['changelog'])
        (cl, lq) = find_changelog(tree, merge=True)
        self.assertEqual(str(cl), cl_block1)
        self.assertEqual(lq, True)

    def test_find_changelog_lq_unversioned_debian_symlink(self):
        # LarstiQ mode, but with an unversioned "debian" -> "." symlink.
        # Bug 619295
        self.requireFeature(SymlinkFeature)
        tree = self.make_branch_and_tree('.')
        self.write_changelog('changelog')
        tree.add(['changelog'])
        os.symlink('.', 'debian')
        self.assertRaises(AddChangelogError, find_changelog, tree, merge=True)

    def test_find_changelog_nomerge_lq(self):
        tree = self.make_branch_and_tree('.')
        self.write_changelog('changelog')
        tree.add(['changelog'])
        self.assertRaises(
            MissingChangelogError, find_changelog, tree, merge=False)

    def test_find_changelog_nochangelog(self):
        tree = self.make_branch_and_tree('.')
        self.write_changelog('changelog')
        self.assertRaises(
            MissingChangelogError, find_changelog, tree, merge=False)

    def test_find_changelog_nochangelog_merge(self):
        tree = self.make_branch_and_tree('.')
        self.assertRaises(
            MissingChangelogError, find_changelog, tree, merge=True)

    def test_find_changelog_symlink(self):
        """When there was a symlink debian -> . then the code used to break"""
        self.requireFeature(SymlinkFeature)
        tree = self.make_branch_and_tree('.')
        self.write_changelog('changelog')
        tree.add(['changelog'])
        os.symlink('.', 'debian')
        tree.add(['debian'])
        (cl, lq) = find_changelog(tree, merge=True)
        self.assertEqual(str(cl), cl_block1)
        self.assertEqual(lq, True)

    def test_find_changelog_symlink_naughty(self):
        tree = self.make_branch_and_tree('.')
        os.mkdir('debian')
        self.write_changelog('debian/changelog')
        with open('changelog', 'w') as f:
            f.write('Naughty, naughty')
        tree.add(['changelog', 'debian', 'debian/changelog'])
        (cl, lq) = find_changelog(tree, merge=True)
        self.assertEqual(str(cl), cl_block1)
        self.assertEqual(lq, False)

    def test_changelog_not_added(self):
        tree = self.make_branch_and_tree('.')
        os.mkdir('debian')
        self.write_changelog('debian/changelog')
        self.assertRaises(AddChangelogError, find_changelog, tree, merge=False)


class TarballNameTests(TestCase):

    def test_tarball_name(self):
        self.assertEqual(
            tarball_name("package", "0.1", None),
            "package_0.1.orig.tar.gz")
        self.assertEqual(
            tarball_name("package", Version("0.1"), None),
            "package_0.1.orig.tar.gz")
        self.assertEqual(
            tarball_name("package", Version("0.1"), None, format='bz2'),
            "package_0.1.orig.tar.bz2")
        self.assertEqual(
            tarball_name("package", Version("0.1"), None, format='xz'),
            "package_0.1.orig.tar.xz")
        self.assertEqual(
            tarball_name("package", Version("0.1"), "la", format='xz'),
            "package_0.1.orig-la.tar.xz")


DistroInfoFeature = ModuleAvailableFeature('distro_info')


class SuiteToDistributionTests(TestCase):

    _test_needs_features = [DistroInfoFeature]

    def _do_lookup(self, target):
        return suite_to_distribution(target)

    def lookup_ubuntu(self, target):
        self.assertEqual(self._do_lookup(target), 'ubuntu')

    def lookup_debian(self, target):
        self.assertEqual(self._do_lookup(target), 'debian')

    def lookup_kali(self, target):
        self.assertEqual(self._do_lookup(target), 'kali')

    def lookup_other(self, target):
        self.assertEqual(self._do_lookup(target), None)

    def test_lookup_ubuntu(self):
        self.lookup_ubuntu('intrepid')
        self.lookup_ubuntu('hardy-proposed')
        self.lookup_ubuntu('gutsy-updates')
        self.lookup_ubuntu('feisty-security')
        self.lookup_ubuntu('dapper-backports')

    def test_lookup_debian(self):
        self.lookup_debian('unstable')
        self.lookup_debian('stable-security')
        self.lookup_debian('testing-proposed-updates')
        self.lookup_debian('etch-backports')

    def test_lookup_kali(self):
        self.lookup_kali('kali-dev')
        self.lookup_kali('kali-rolling')
        self.lookup_kali('kali')

    def test_lookup_other(self):
        self.lookup_other('not-a-target')
        self.lookup_other("debian")
        self.lookup_other("ubuntu")


class LookupDistributionTests(SuiteToDistributionTests):

    _test_needs_features = [DistroInfoFeature]

    def _do_lookup(self, target):
        return lookup_distribution(target)

    def test_lookup_other(self):
        self.lookup_other('not-a-target')
        self.lookup_debian("debian")
        self.lookup_ubuntu("ubuntu")
        self.lookup_ubuntu("Ubuntu")


class MoveFileTests(TestCaseInTempDir):

    def test_move_file_non_extant(self):
        self.build_tree(['a'])
        move_file_if_different('a', 'b', None)
        self.assertPathDoesNotExist('a')
        self.assertPathExists('b')

    def test_move_file_samefile(self):
        self.build_tree(['a'])
        move_file_if_different('a', 'a', None)
        self.assertPathExists('a')

    def test_move_file_same_md5(self):
        self.build_tree(['a'])
        md5sum = hashlib.md5()
        with open('a', 'rb') as f:
            md5sum.update(f.read())
        shutil.copy('a', 'b')
        move_file_if_different('a', 'b', md5sum.hexdigest())
        self.assertPathExists('a')
        self.assertPathExists('b')

    def test_move_file_diff_md5(self):
        self.build_tree(['a', 'b'])
        md5sum = hashlib.md5()
        with open('a', 'rb') as f:
            md5sum.update(f.read())
        a_hexdigest = md5sum.hexdigest()
        md5sum = hashlib.md5()
        with open('b', 'rb') as f:
            md5sum.update(f.read())
        b_hexdigest = md5sum.hexdigest()
        self.assertNotEqual(a_hexdigest, b_hexdigest)
        move_file_if_different('a', 'b', a_hexdigest)
        self.assertPathDoesNotExist('a')
        self.assertPathExists('b')
        md5sum = hashlib.md5()
        with open('b', 'rb') as f:
            md5sum.update(f.read())
        self.assertEqual(md5sum.hexdigest(), a_hexdigest)


class WriteFileTests(TestCaseInTempDir):

    def test_write_non_extant(self):
        write_if_different(b"foo", 'a')
        self.assertPathExists('a')
        self.check_file_contents('a', b"foo")

    def test_write_file_same(self):
        write_if_different(b"foo", 'a')
        self.assertPathExists('a')
        self.check_file_contents('a', b"foo")
        write_if_different(b"foo", 'a')
        self.assertPathExists('a')
        self.check_file_contents('a', b"foo")

    def test_write_file_different(self):
        write_if_different(b"foo", 'a')
        self.assertPathExists('a')
        self.check_file_contents('a', b"foo")
        write_if_different(b"bar", 'a')
        self.assertPathExists('a')
        self.check_file_contents('a', b"bar")


class DgetTests(TestCaseWithTransport):

    def test_dget_local(self):
        builder = SourcePackageBuilder("package", Version("0.1-1"))
        builder.add_upstream_file("foo")
        builder.add_default_control()
        builder.build()
        self.build_tree(["target/"])
        dget(builder.dsc_name(), 'target')
        self.assertPathExists(os.path.join("target", builder.dsc_name()))
        self.assertPathExists(os.path.join("target", builder.tar_name()))
        self.assertPathExists(os.path.join("target", builder.diff_name()))

    def test_dget_transport(self):
        builder = SourcePackageBuilder("package", Version("0.1-1"))
        builder.add_upstream_file("foo")
        builder.add_default_control()
        builder.build()
        self.build_tree(["target/"])
        dget(self.get_url(builder.dsc_name()), 'target')
        self.assertPathExists(os.path.join("target", builder.dsc_name()))
        self.assertPathExists(os.path.join("target", builder.tar_name()))
        self.assertPathExists(os.path.join("target", builder.diff_name()))

    def test_dget_missing_dsc(self):
        builder = SourcePackageBuilder("package", Version("0.1-1"))
        builder.add_upstream_file("foo")
        builder.add_default_control()
        # No builder.build()
        self.build_tree(["target/"])
        self.assertRaises(
            bzr_errors.NoSuchFile, dget,
            self.get_url(builder.dsc_name()), 'target')

    def test_dget_missing_file(self):
        builder = SourcePackageBuilder("package", Version("0.1-1"))
        builder.add_upstream_file("foo")
        builder.add_default_control()
        builder.build()
        os.unlink(builder.tar_name())
        self.build_tree(["target/"])
        self.assertRaises(
            bzr_errors.NoSuchFile, dget,
            self.get_url(builder.dsc_name()), 'target')

    def test_dget_missing_target(self):
        builder = SourcePackageBuilder("package", Version("0.1-1"))
        builder.add_upstream_file("foo")
        builder.add_default_control()
        builder.build()
        self.assertRaises(
            bzr_errors.NotADirectory, dget,
            self.get_url(builder.dsc_name()), 'target')

    def test_dget_changes(self):
        builder = SourcePackageBuilder("package", Version("0.1-1"))
        builder.add_upstream_file("foo")
        builder.add_default_control()
        builder.build()
        self.build_tree(["target/"])
        dget_changes(builder.changes_name(), 'target')
        self.assertPathExists(os.path.join("target", builder.dsc_name()))
        self.assertPathExists(os.path.join("target", builder.tar_name()))
        self.assertPathExists(os.path.join("target", builder.diff_name()))
        self.assertPathExists(os.path.join("target", builder.changes_name()))


class ParentDirTests(TestCase):

    def test_get_parent_dir(self):
        self.assertEqual(get_parent_dir("a"), '')
        self.assertEqual(get_parent_dir("a/"), '')
        self.assertEqual(get_parent_dir("a/b"), 'a')
        self.assertEqual(get_parent_dir("a/b/"), 'a')
        self.assertEqual(get_parent_dir("a/b/c"), 'a/b')


class ChangelogInfoTests(TestCaseWithTransport):

    def test_find_extra_authors_none(self):
        changes = ["  * Do foo", "  * Do bar"]
        authors = find_extra_authors(changes)
        self.assertEqual([], authors)

    def test_find_extra_authors(self):
        changes = ["  * Do foo", "", "  [ A. Hacker ]", "  * Do bar", "",
                   "  [ B. Hacker ]", "  [ A. Hacker}"]
        authors = find_extra_authors(changes)
        self.assertEqual([u"A. Hacker", u"B. Hacker"], authors)
        self.assertEqual([text_type]*len(authors), list(map(type, authors)))

    def test_find_extra_authors_utf8(self):
        changes = [u"  * Do foo", u"", "  [ \xe1. Hacker ]", "  * Do bar", "",
                   u"  [ \xe7. Hacker ]", "  [ A. Hacker}"]
        authors = find_extra_authors(changes)
        self.assertEqual([u"\xe1. Hacker", u"\xe7. Hacker"], authors)
        self.assertEqual([text_type]*len(authors), list(map(type, authors)))

    def test_find_extra_authors_iso_8859_1(self):
        # We try to treat lines as utf-8, but if that fails to decode, we fall
        # back to iso-8859-1
        changes = ["  * Do foo", "", "  [ \xe1. Hacker ]", "  * Do bar", "",
                   "  [ \xe7. Hacker ]", "  [ A. Hacker}"]
        authors = find_extra_authors(changes)
        self.assertEqual([u"\xe1. Hacker", u"\xe7. Hacker"], authors)
        self.assertEqual([text_type]*len(authors), list(map(type, authors)))

    def test_find_extra_authors_no_changes(self):
        authors = find_extra_authors([])
        self.assertEqual([], authors)

    def assert_thanks_is(self, changes, expected_thanks):
        thanks = find_thanks(changes)
        self.assertEqual(expected_thanks, thanks)
        self.assertEqual([text_type]*len(thanks), list(map(type, thanks)))

    def test_find_thanks_no_changes(self):
        self.assert_thanks_is([], [])

    def test_find_thanks_none(self):
        changes = ["  * Do foo", "  * Do bar"]
        self.assert_thanks_is(changes, [])

    def test_find_thanks(self):
        changes = ["  * Thanks to A. Hacker"]
        self.assert_thanks_is(changes, [u"A. Hacker"])
        changes = ["  * Thanks to James A. Hacker"]
        self.assert_thanks_is(changes, [u"James A. Hacker"])
        changes = ["  * Thankyou to B. Hacker"]
        self.assert_thanks_is(changes, [u"B. Hacker"])
        changes = ["  * thanks to A. Hacker"]
        self.assert_thanks_is(changes, [u"A. Hacker"])
        changes = ["  * thankyou to B. Hacker"]
        self.assert_thanks_is(changes, [u"B. Hacker"])
        changes = ["  * Thanks A. Hacker"]
        self.assert_thanks_is(changes, [u"A. Hacker"])
        changes = ["  * Thankyou B.  Hacker"]
        self.assert_thanks_is(changes, [u"B. Hacker"])
        changes = ["  * Thanks to Mark A. Super-Hacker"]
        self.assert_thanks_is(changes, [u"Mark A. Super-Hacker"])
        changes = ["  * Thanks to A. Hacker <ahacker@example.com>"]
        self.assert_thanks_is(changes, [u"A. Hacker <ahacker@example.com>"])
        changes = [u"  * Thanks to Adeodato Sim\xc3\xb3"]
        self.assert_thanks_is(changes, [u"Adeodato Sim\xc3\xb3"])
        changes = [u"  * Thanks to \xc1deodato Sim\xc3\xb3"]
        self.assert_thanks_is(changes, [u"\xc1deodato Sim\xc3\xb3"])

    def test_find_bugs_fixed_no_changes(self):
        self.assertEqual([], find_bugs_fixed([], None, _lplib=MockLaunchpad()))

    def test_find_bugs_fixed_none(self):
        changes = ["  * Do foo", "  * Do bar"]
        bugs = find_bugs_fixed(changes, None, _lplib=MockLaunchpad())
        self.assertEqual([], bugs)

    def test_find_bugs_fixed_debian(self):
        wt = self.make_branch_and_tree(".")
        changes = ["  * Closes: #12345, 56789", "  * closes:bug45678"]
        bugs = find_bugs_fixed(changes, wt.branch, _lplib=MockLaunchpad())
        self.assertEqual(
            ["http://bugs.debian.org/12345 fixed",
             "http://bugs.debian.org/56789 fixed",
             "http://bugs.debian.org/45678 fixed"], bugs)

    def test_find_bugs_fixed_debian_with_ubuntu_links(self):
        wt = self.make_branch_and_tree(".")
        changes = ["  * Closes: #12345", "  * closes:bug45678"]
        lplib = MockLaunchpad(
            debian_bug_to_ubuntu_bugs={
                "12345": ("998877", "987654"),
                "45678": ("87654",)})
        bugs = find_bugs_fixed(changes, wt.branch, _lplib=lplib)
        self.assertEqual([], lplib.ubuntu_bug_lookups)
        self.assertEqual(["12345", "45678"], lplib.debian_bug_lookups)
        self.assertEqual(
            ["http://bugs.debian.org/12345 fixed",
             "http://bugs.debian.org/45678 fixed",
             "https://launchpad.net/bugs/87654 fixed"], bugs)

    def test_find_bugs_fixed_lp(self):
        wt = self.make_branch_and_tree(".")
        changes = ["  * LP: #12345,#56789", "  * lp:  #45678"]
        bugs = find_bugs_fixed(changes, wt.branch, _lplib=MockLaunchpad())
        self.assertEqual(
            ["https://launchpad.net/bugs/12345 fixed",
             "https://launchpad.net/bugs/56789 fixed",
             "https://launchpad.net/bugs/45678 fixed"], bugs)

    def test_find_bugs_fixed_lp_with_debian_links(self):
        wt = self.make_branch_and_tree(".")
        changes = ["  * LP: #12345", "  * lp:  #45678"]
        lplib = MockLaunchpad(
            ubuntu_bug_to_debian_bugs={
                "12345": ("998877", "987654"), "45678": ("87654",)})
        bugs = find_bugs_fixed(changes, wt.branch, _lplib=lplib)
        self.assertEqual([], lplib.debian_bug_lookups)
        self.assertEqual(["12345", "45678"], lplib.ubuntu_bug_lookups)
        self.assertEqual(
            ["https://launchpad.net/bugs/12345 fixed",
             "https://launchpad.net/bugs/45678 fixed",
             "http://bugs.debian.org/87654 fixed"], bugs)

    def test_get_commit_info_none(self):
        wt = self.make_branch_and_tree(".")
        changelog = Changelog()
        message, authors, thanks, bugs = get_commit_info_from_changelog(
            changelog, wt.branch, _lplib=MockLaunchpad())
        self.assertEqual(None, message)
        self.assertEqual([], authors)
        self.assertEqual([], thanks)
        self.assertEqual([], bugs)

    def test_get_commit_message_info(self):
        wt = self.make_branch_and_tree(".")
        changelog = Changelog()
        changes = ["  [ A. Hacker ]", "  * First change, LP: #12345",
                   "  * Second change, thanks to B. Hacker"]
        author = "J. Maintainer <maint@example.com"
        changelog.new_block(changes=changes, author=author)
        message, authors, thanks, bugs = get_commit_info_from_changelog(
            changelog, wt.branch, _lplib=MockLaunchpad())
        self.assertEqual("\n".join(strip_changelog_message(changes)), message)
        self.assertEqual([author]+find_extra_authors(changes), authors)
        self.assertEqual(text_type, type(authors[0]))
        self.assertEqual(find_thanks(changes), thanks)
        self.assertEqual(find_bugs_fixed(
            changes, wt.branch, _lplib=MockLaunchpad()), bugs)

    def assertUnicodeCommitInfo(self, changes):
        wt = self.make_branch_and_tree(".")
        changelog = Changelog()
        author = "J. Maintainer <maint@example.com>"
        changelog.new_block(changes=changes, author=author)
        message, authors, thanks, bugs = get_commit_info_from_changelog(
            changelog, wt.branch, _lplib=MockLaunchpad())
        self.assertEqual(u'[ \xc1. Hacker ]\n'
                         u'* First ch\xe1nge, LP: #12345\n'
                         u'* Second change, thanks to \xde. Hacker',
                         message)
        self.assertEqual([author, u'\xc1. Hacker'], authors)
        self.assertEqual(text_type, type(authors[0]))
        self.assertEqual([u'\xde. Hacker'], thanks)
        self.assertEqual(['https://launchpad.net/bugs/12345 fixed'], bugs)

    def test_get_commit_info_unicode(self):
        changes = [u"  [ \xc1. Hacker ]",
                   u"  * First ch\xe1nge, LP: #12345",
                   u"  * Second change, thanks to \xde. Hacker"]
        self.assertUnicodeCommitInfo(changes)


class MockLaunchpad(object):

    def __init__(self, debian_bug_to_ubuntu_bugs={},
                 ubuntu_bug_to_debian_bugs={}):
        self.debian_bug_to_ubuntu_bugs = debian_bug_to_ubuntu_bugs
        self.ubuntu_bug_to_debian_bugs = ubuntu_bug_to_debian_bugs
        self.debian_bug_lookups = []
        self.ubuntu_bug_lookups = []

    def ubuntu_bugs_for_debian_bug(self, debian_bug):
        self.debian_bug_lookups.append(debian_bug)
        try:
            return self.debian_bug_to_ubuntu_bugs[debian_bug]
        except KeyError:
            return []

    def debian_bugs_for_ubuntu_bug(self, ubuntu_bug):
        self.ubuntu_bug_lookups.append(ubuntu_bug)
        try:
            return self.ubuntu_bug_to_debian_bugs[ubuntu_bug]
        except KeyError:
            return []


class FindPreviousUploadTests(TestCase):

    def make_changelog(self, versions_and_distributions):
        cl = Changelog()
        changes = ["  [ A. Hacker ]", "  * Something"]
        author = "J. Maintainer <maint@example.com>"
        for version, distro in versions_and_distributions:
            cl.new_block(changes=changes, author=author,
                         distributions=distro, version=version)
        return cl

    def test_find_previous_upload_debian(self):
        cl = self.make_changelog(
            [("0.1-1", "unstable"),
             ("0.1-2", "unstable")])
        self.assertEqual(Version("0.1-1"), changelog_find_previous_upload(cl))
        cl = self.make_changelog(
            [("0.1-1", "unstable"),
             ("0.1-1.1", "stable-security"), ("0.1-2", "unstable")])
        self.assertEqual(Version("0.1-1"), changelog_find_previous_upload(cl))

    def test_find_previous_upload_ubuntu(self):
        cl = self.make_changelog(
            [("0.1-1", "lucid"),
             ("0.1-2", "lucid")])
        self.assertEqual(Version("0.1-1"), changelog_find_previous_upload(cl))
        cl = self.make_changelog(
            [("0.1-1", "lucid"),
             ("0.1-1.1", "unstable"), ("0.1-2", "maverick")])
        self.requireFeature(DistroInfoFeature)
        self.assertEqual(
                Version("0.1-1"), changelog_find_previous_upload(cl))

    def test_find_previous_upload_ubuntu_pocket(self):
        cl = self.make_changelog(
            [("0.1-1", "lucid-updates"),
             ("0.1-2", "lucid-updates")])
        self.assertEqual(Version("0.1-1"), changelog_find_previous_upload(cl))

    def test_find_previous_upload_unknown(self):
        cl = self.make_changelog(
            [("0.1-1", "lucid"),
             ("0.1-2", "dunno")])
        self.assertRaises(NoPreviousUpload, changelog_find_previous_upload, cl)

    def test_find_previous_upload_missing(self):
        cl = self.make_changelog(
            [("0.1-1", "unstable"),
             ("0.1-2", "lucid")])
        self.assertRaises(NoPreviousUpload, changelog_find_previous_upload, cl)
        cl = self.make_changelog([("0.1-1", "unstable")])
        self.assertRaises(NoPreviousUpload, changelog_find_previous_upload, cl)

    def test_find_previous_upload_unreleased(self):
        cl = self.make_changelog(
            [("0.1-1", "unstable"),
             ("0.1-2", "UNRELEASED")])
        self.assertEqual(Version("0.1-1"), changelog_find_previous_upload(cl))


class SourceFormatTests(TestCaseWithTransport):

    def test_no_source_format_file(self):
        tree = self.make_branch_and_tree('.')
        self.assertEquals("1.0", tree_get_source_format(tree))

    def test_source_format_newline(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents(
            [("debian/", ), ("debian/source/",),
             ("debian/source/format", "3.0 (native)\n")])
        tree.add(["debian", "debian/source", "debian/source/format"])
        self.assertEquals("3.0 (native)", tree_get_source_format(tree))

    def test_source_format(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents(
            [("debian/",), ("debian/source/",),
             ("debian/source/format", "3.0 (quilt)")])
        tree.add(["debian", "debian/source", "debian/source/format"])
        self.assertEquals("3.0 (quilt)", tree_get_source_format(tree))

    def test_source_format_file_unversioned(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents(
            [("debian/",), ("debian/source/",),
             ("debian/source/format", "3.0 (quilt)")])
        self.assertEquals("3.0 (quilt)", tree_get_source_format(tree))


class GuessBuildTypeTests(TestCaseWithTransport):
    """Tests for guess_build_type."""

    def writeVersionFile(self, tree, format_string):
        """Write a Debian source format file.

        :param tree: Tree to write to.
        :param format_string: Format string to write.
        """
        self.build_tree_contents(
            [("debian/",), ("debian/source/",),
             ("debian/source/format", format_string)])
        tree.add(["debian", "debian/source", "debian/source/format"])

    def test_normal_source_format(self):
        # Normal source format -> NORMAL
        tree = self.make_branch_and_tree('.')
        self.writeVersionFile(tree, "3.0 (quilt)")
        self.assertEquals(
            BUILD_TYPE_NORMAL,
            guess_build_type(tree, None, contains_upstream_source=True))

    def test_normal_source_format_merge(self):
        # Normal source format without upstream source -> MERGE
        tree = self.make_branch_and_tree('.')
        self.writeVersionFile(tree, "3.0 (quilt)")
        self.assertEquals(
            BUILD_TYPE_MERGE,
            guess_build_type(tree, None, contains_upstream_source=False))

    def test_native_source_format(self):
        # Native source format -> NATIVE
        tree = self.make_branch_and_tree('.')
        self.writeVersionFile(tree, "3.0 (native)")
        self.assertEquals(
            BUILD_TYPE_NATIVE,
            guess_build_type(tree, None, contains_upstream_source=True))

    def test_prev_version_native(self):
        # Native package version -> NATIVE
        tree = self.make_branch_and_tree('.')
        self.assertEquals(
            BUILD_TYPE_NATIVE,
            guess_build_type(
                tree, Version("1.0"), contains_upstream_source=True))

    def test_empty(self):
        # Empty tree and a non-native package -> NORMAL
        tree = self.make_branch_and_tree('.')
        self.assertEquals(
            BUILD_TYPE_NORMAL,
            guess_build_type(
                tree, Version("1.0-1"), contains_upstream_source=None))

    def test_no_upstream_source(self):
        # No upstream source code and a non-native package -> MERGE
        tree = self.make_branch_and_tree('.')
        tree.mkdir("debian")
        self.assertEquals(
            BUILD_TYPE_MERGE,
            guess_build_type(
                tree, Version("1.0-1"), contains_upstream_source=False))

    def test_default(self):
        # Upstream source code and a non-native package -> NORMAL
        tree = self.make_branch_and_tree('.')
        self.assertEquals(
            BUILD_TYPE_NORMAL,
            guess_build_type(
                tree, Version("1.0-1"), contains_upstream_source=True))

    def test_inconsistent(self):
        # If version string and source format disagree on whether the package
        # is native, raise an exception.
        tree = self.make_branch_and_tree('.')
        self.writeVersionFile(tree, "3.0 (quilt)")
        e = self.assertRaises(
            InconsistentSourceFormatError, guess_build_type, tree,
            Version("1.0"), contains_upstream_source=True)
        self.assertEquals(
            "Inconsistency between source format and version: "
            "version is native, format is not native.", str(e))


class TestExtractOrigTarballs(TestCaseInTempDir):

    def create_tarball(self, package, version, compression, part=None):
        basedir = "%s-%s" % (package, version)
        os.mkdir(basedir)
        try:
            f = open(os.path.join(basedir, "README"), 'w')
            try:
                f.write("Hi\n")
            finally:
                f.close()
            prefix = "%s_%s.orig" % (package, version)
            if part is not None:
                prefix += "-%s" % part
            tar_path = os.path.abspath(prefix + ".tar." + compression)
            if compression == "gz":
                f = gzip.GzipFile(tar_path, "w")
            elif compression == "bz2":
                f = bz2.BZ2File(tar_path, "w")
            elif compression == "xz":
                import lzma
                f = lzma.LZMAFile(tar_path, "w")
            else:
                raise AssertionError(
                    "Unknown compressin type %r" % compression)
            try:
                tf = tarfile.open(None, 'w', f)
                try:
                    tf.add(basedir)
                finally:
                    tf.close()
            finally:
                f.close()
        finally:
            shutil.rmtree(basedir)
        return tar_path

    def test_single_orig_tar_gz(self):
        tar_path = self.create_tarball("package", "0.1", "gz")
        os.mkdir("target")
        extract_orig_tarballs(
            [(tar_path, None)], "target", strip_components=1)
        self.assertEquals(os.listdir("target"), ["README"])

    def test_single_orig_tar_bz2(self):
        tar_path = self.create_tarball("package", "0.1", "bz2")
        os.mkdir("target")
        extract_orig_tarballs(
            [(tar_path, None)], "target", strip_components=1)
        self.assertEquals(os.listdir("target"), ["README"])

    def test_single_orig_tar_xz(self):
        self.requireFeature(LzmaFeature)
        tar_path = self.create_tarball("package", "0.1", "xz")
        os.mkdir("target")
        extract_orig_tarballs(
            [(tar_path, None)], "target", strip_components=1)
        self.assertEquals(os.listdir("target"), ["README"])

    def test_multiple_tarballs(self):
        base_tar_path = self.create_tarball("package", "0.1", "bz2")
        tar_path_extra = self.create_tarball(
            "package", "0.1", "bz2", part="extra")
        os.mkdir("target")
        extract_orig_tarballs(
            [(base_tar_path, None), (tar_path_extra, "extra")], "target",
            strip_components=1)
        self.assertEquals(
            sorted(os.listdir("target")),
            sorted(["README", "extra"]))


class ComponentFromOrigTarballTests(TestCase):

    def test_base_tarball(self):
        self.assertIs(
            None,
            component_from_orig_tarball(
                "foo_0.1.orig.tar.gz", "foo", "0.1"))
        self.assertRaises(
            ValueError,
            component_from_orig_tarball, "foo_0.1.orig.tar.gz", "bar", "0.1")

    def test_invalid_extension(self):
        self.assertRaises(
            ValueError,
            component_from_orig_tarball, "foo_0.1.orig.unknown", "foo", "0.1")

    def test_component(self):
        self.assertEquals(
            "comp",
            component_from_orig_tarball(
                "foo_0.1.orig-comp.tar.gz", "foo", "0.1"))
        self.assertEquals(
            "comp-dash",
            component_from_orig_tarball(
                "foo_0.1.orig-comp-dash.tar.gz", "foo", "0.1"))

    def test_invalid_character(self):
        self.assertRaises(
            ValueError,
            component_from_orig_tarball, "foo_0.1.orig;.tar.gz", "foo", "0.1")


class TreeContainsUpstreamSourceTests(TestCaseWithTransport):

    def test_empty(self):
        tree = self.make_branch_and_tree('.')
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertIs(None, tree_contains_upstream_source(tree))

    def test_debian_dir_only(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree(['debian/'])
        tree.add(['debian'])
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertFalse(tree_contains_upstream_source(tree))

    def test_debian_dir_and_bzr_builddeb(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree(['debian/', '.bzr-builddeb/'])
        tree.add(['debian', '.bzr-builddeb'])
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertFalse(tree_contains_upstream_source(tree))

    def test_with_upstream_source(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree(['debian/', 'src/'])
        tree.add(['debian', 'src'])
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertTrue(tree_contains_upstream_source(tree))

    def test_with_unversioned_extra_data(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree(['debian/', 'x'])
        tree.add(['debian'])
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertFalse(tree_contains_upstream_source(tree))


class BuildArchitectureTests(TestCase):

    def test_is_str(self):
        self.assertIsInstance(get_build_architecture(), text_type)


class FilesExcludedTests(TestCaseWithTransport):

    def test_file_missing(self):
        tree = self.make_branch_and_tree('.')
        self.assertRaises(bzr_errors.NoSuchFile, get_files_excluded, tree)

    def test_not_set(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/copyright', """\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: blah

Files: *
Copyright:
 (c) Somebody
License: MIT
""")])
        tree.add(['debian', 'debian/copyright'])
        self.assertEqual([], get_files_excluded(tree))

    def test_set(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('debian/', ),
            ('debian/copyright', """\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: blah
Files-Excluded: blah/* flattr.png

Files: *
Copyright:
 (c) Somebody
License: MIT
""")])
        tree.add(['debian', 'debian/copyright'])
        self.assertEqual(['blah/*', 'flattr.png'], get_files_excluded(tree))



#    hooks.py -- Hook support for builddeb.
#    Copyright (C) 2006 James Westby <jw+debian@jameswestby.net>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from __future__ import absolute_import

import subprocess

from ...errors import BzrError
from ...trace import note


class HookFailedError(BzrError):
    _fmt = 'The "%(hook_name)s" hook failed.'

    def __init__(self, hook_name):
        BzrError.__init__(self, hook_name=hook_name)


def run_hook(tree, hook_name, config, wd="."):
    hook = config.get_hook(hook_name)
    if hook is None:
        return
    note("Running %s as %s hook" % (hook, hook_name))
    proc = subprocess.Popen(hook, shell=True, cwd=tree.abspath(wd))
    proc.wait()
    if proc.returncode != 0:
        raise HookFailedError(hook_name)



#    upstream/branch.py -- Upstream branch source provider
#    Copyright (C) 2010-2011 Canonical Ltd.
#    Copyright (C) 2009 Jelmer Vernooij
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from debian.changelog import Version
import os
import re
import subprocess
import tempfile

from debmutate.versions import git_snapshot_data_from_version

from .... import osutils
from ....branch import (
    Branch,
    BranchWriteLockResult,
    )
from ....errors import (
    BzrError,
    GhostRevisionsHaveNoRevno,
    InvalidRevisionId,
    NoSuchRevision,
    NoSuchTag,
    RevisionNotPresent,
    UnsupportedOperation,
    )
from ..repack_tarball import get_filetype, repack_tarball
from ....revision import NULL_REVISION
from ....revisionspec import RevisionSpec
from ....trace import note, mutter
from ....tree import Tree

try:
    from ....revisionspec import InvalidRevisionSpec
except ImportError:  # Breezy < 3.2
    from ....errors import InvalidRevisionSpec

from ..errors import (
    MultipleUpstreamTarballsNotSupported,
    )
from .. import gettext
from . import (
    UpstreamSource,
    PackageVersionNotPresent,
    new_tarball_name,
    )
from ....export import (
    export,
    )


def upstream_tag_to_version(tag_name, package=None):
    """Take a tag name and return the upstream version, or None."""
    if tag_name.endswith('-release'):
        tag_name = tag_name[:-len('-release')]
    if tag_name.startswith("release-"):
        tag_name = tag_name[len("release-"):]
    if tag_name[0] == "v" and tag_name[1].isdigit():
        tag_name = tag_name[1:]
    if (package is not None and (
          tag_name.startswith("%s-" % package) or
          tag_name.startswith("%s_" % package))):
        tag_name = tag_name[len(package)+1:]
    if '_' in tag_name and '.' not in tag_name:
        tag_name = tag_name.replace('_', '.')
    if tag_name.count('_') == 1 and tag_name.startswith('0.'):
        # This is a style commonly used for perl packages.
        # Most debian packages seem to just drop the underscore.
        tag_name = tag_name.replace('_', '')
    if all([c.isdigit() or c in (".", "~") for c in tag_name]):
        return tag_name
    return None


def _upstream_branch_version(
        revhistory, upstream_revision, reverse_tag_dict, package,
        previous_version, add_rev):
    """Determine the version string of an upstream branch.

    The upstream version is determined from the most recent tag
    in the upstream branch. If that tag does not point at the last revision,
    the revision number is added to it (<version>+bzr<revno>).

    If there are no tags set on the upstream branch, the previous Debian
    version is used and combined with the bzr revision number
    (usually <version>+bzr<revno>).

    :param revhistory: Reverse branch revision history.
    :param reverse_tag_dict: Reverse tag dictionary (revid -> list of tags)
    :param package: Name of package.
    :param previous_version: Previous upstream version in debian changelog.
    :param add_rev: Function that can add a revision suffix to a version
        string.
    :return: Name of the upstream revision.
    """
    if upstream_revision == NULL_REVISION:
        # No new version to merge
        return previous_version
    try:
        for r in revhistory:
            if r in reverse_tag_dict:
                # If there is a newer version tagged in branch,
                # convert to upstream version
                # return <upstream_version>+bzr<revno>
                for tag in reverse_tag_dict[r]:
                    upstream_version = upstream_tag_to_version(
                            tag, package=package)
                    if upstream_version is not None:
                        if r != upstream_revision:
                            upstream_version = add_rev(
                                str(upstream_version), upstream_revision)
                        return upstream_version
    except RevisionNotPresent:
        # Ghost revision somewhere on mainline.
        pass
    if previous_version is None:
        return None
    return add_rev(str(previous_version), upstream_revision)


def extract_gitid(rev):
    """Extract the git SHA of a revision from a Revision object.

    :param rev: Revision object
    :return: 40 character Git SHA as bytes,
        None if this was not a git revision or
         if the git sha could not be determined (dulwich not available).
    """
    try:
        from ....git import extract_git_foreign_revid
    except ImportError:
        # No git support
        return None
    else:
        try:
            return extract_git_foreign_revid(rev)
        except InvalidRevisionId:
            return None


def extract_svn_revno(rev):
    """Extract the Subversion number of a revision from a Revision object.

    :param rev: Revision object
    :return: Revision number, None if this was not a Subversion revision or
         if the revision number could not be determined
         (bzr-svn not available).
    """
    try:
        from ...svn import extract_svn_foreign_revid
    except ImportError:
        # No svn support
        return None
    else:
        try:
            (svn_uuid, branch_path, svn_revno) = extract_svn_foreign_revid(rev)
        except InvalidRevisionId:
            return None
        else:
            return svn_revno


def upstream_version_add_revision(
        upstream_branch, version_string, revid, sep='+'):
    """Update the revision in a upstream version string.

    :param branch: Branch in which the revision can be found
    :param version_string: Original version string
    :param revid: Revision id of the revision
    :param sep: Separator to use when adding snapshot
    """
    try:
        revno = upstream_branch.revision_id_to_dotted_revno(revid)
    except GhostRevisionsHaveNoRevno:
        pass
    else:
        revno_str = '.'.join(map(str, revno))

        m = re.match(r"^(.*)([\+~])bzr(\d+)$", version_string)
        if m:
            return "%s%sbzr%s" % (m.group(1), m.group(2), revno_str)

    rev = upstream_branch.repository.get_revision(revid)
    gitid = extract_gitid(rev)
    if gitid:
        gitid = gitid[:7].decode('ascii')
        gitdate = osutils.format_date(
            rev.timestamp, rev.timezone, date_fmt='%Y%m%d', show_offset=False)

    m = re.match(r"^(.*)([\+~-])git(\d{8})\.([a-f0-9]{7})$", version_string)
    if m and gitid:
        return "%s%sgit%s.%s" % (m.group(1), m.group(2), gitdate, gitid)

    m = re.match(r"^(.*)([\+~-])git(\d{8})\.(\d+)\.([a-f0-9]{7})$",
                 version_string)
    if m and gitid:
        if gitdate == m.group(3):
            snapshot = int(m.group(4)) + 1
        else:
            snapshot = 0
        return "%s%sgit%s.%d.%s" % (
            m.group(1), m.group(2), gitdate, snapshot, gitid)

    m = re.match(r"^(.*)([\+~-])git(\d{8})$", version_string)
    if m and gitid:
        return "%s%sgit%s" % (m.group(1), m.group(2), gitdate)

    svn_revno = extract_svn_revno(rev)

    m = re.match(r"^(.*)([\+~])svn(\d+)$", version_string)
    # FIXME: Raise error if +svn/~svn is present and svn_revno is not set?
    if m and svn_revno:
        return "%s%ssvn%d" % (m.group(1), m.group(2), svn_revno)

    if svn_revno:
        return "%s%ssvn%d" % (version_string, sep, svn_revno)
    elif gitid:
        return "%s%sgit%s.%s" % (version_string, sep, gitdate, gitid)
    else:
        return "%s%sbzr%s" % (version_string, sep, revno_str)


def get_snapshot_revision(upstream_version):
    """Return the upstream revision specifier if specified in the upstream
    version.

    When packaging an upstream snapshot some people use +vcsnn or ~vcsnn to
    indicate what revision number of the upstream VCS was taken for the
    snapshot. This given an upstream version number this function will return
    an identifier of the upstream revision if it appears to be a snapshot. The
    identifier is a string containing a bzr revision spec, so it can be
    transformed in to a revision.

    :param upstream_version: a string containing the upstream version number.
    :return: a string containing a revision specifier for the revision of the
        upstream branch that the snapshot was taken from, or None if it
        doesn't appear to be a snapshot.
    """
    match = re.search("(?:~|\\+)bzr([0-9]+)$", upstream_version)
    if match is not None:
        return match.groups()[0]
    match = re.search("(?:~|\\+)svn([0-9]+)$", upstream_version)
    if match is not None:
        return "svn:%s" % match.groups()[0]
    match = re.match(r"^(.*)([\+~])git(\d{8})\.([a-f0-9]{7})$",
                     upstream_version)
    if match:
        return "git:%s" % match.group(4)
    match = re.match(r"^(.*)([\+~])git(\d{8})$", upstream_version)
    if match:
        return "date:%s" % match.group(3)
    return None


def upstream_branch_version(upstream_branch, upstream_revision, package,
                            previous_version=None):
    """Determine the version string for a revision in an upstream branch.

    :param upstream_branch: The upstream branch object
    :param upstream_revision: The revision id of the upstream revision
    :param package: The name of the package
    :param previous_version: The previous upstream version string (optional)
    :return: Upstream version string for `upstream_revision`.
    """
    graph = upstream_branch.repository.get_graph()
    if previous_version is not None:
        previous_revision = get_snapshot_revision(previous_version)
    else:
        previous_revision = None
    stop_revids = None
    if previous_revision is not None:
        previous_revspec = RevisionSpec.from_string(previous_revision)
        try:
            previous_revno, previous_revid = previous_revspec.in_history(
                upstream_branch)
        except InvalidRevisionSpec as e:
            # Odd - the revision mentioned in the old version doesn't exist.
            mutter('Unable to find old upstream version %s (%s): %s',
                   previous_version, previous_revision, e)
        else:
            # Trim revision history - we don't care about any revisions
            # before the revision of the previous version
            stop_revids = [previous_revid]
    revhistory = graph.iter_lefthand_ancestry(upstream_revision, stop_revids)
    return _upstream_branch_version(
            revhistory, upstream_revision,
            upstream_branch.tags.get_reverse_tag_dict(), package,
            previous_version,
            lambda version, revision: upstream_version_add_revision(
                upstream_branch, version, revision))


def get_export_upstream_revision(config=None, version=None):
    """Find the revision to use when exporting the upstream source.

    :param config: Config object
    :param version: Optional upstream version to find revision for, if not the
        latest.
    :return: Revision id
    """
    rev = None
    if version is not None:
        rev = get_snapshot_revision(version)
    if rev is None and config is not None:
        rev = config._get_best_opt('export-upstream-revision')
        if rev is not None and version is not None:
            rev = rev.replace('$UPSTREAM_VERSION', version)
    return rev


def guess_upstream_revspec(package, version):
    """Guess revspecs matching an upstream version string."""
    if version.endswith('+ds'):
        version = str(version)[:-len('+ds')]
    if "+bzr" in version or "~bzr" in version:
        yield "revno:%s" % re.match(".*[~+]bzr(\\d+).*", version).group(1)
    (git_id, git_date) = git_snapshot_data_from_version(version)
    if git_id:
        yield "git:%s" % git_id
    if git_date:
        yield "date:%s" % git_date
    yield 'tag:%s' % version
    for prefix in ['rust-']:
        if package.startswith(prefix):
            yield 'tag:%s-%s' % (package[len(prefix):], version)
    yield 'tag:%s-%s' % (package, version)
    yield 'tag:v%s' % version
    yield 'tag:v.%s' % version


try:
    from breezy.memorybranch import MemoryBranch
except ImportError:  # breezy < 3.1.1
    from ....bzr import branch as bzr_branch
    from ....lock import _RelockDebugMixin, LogicalLockResult

    class MemoryBranch(bzr_branch.Branch, _RelockDebugMixin):

        def __init__(self, repository, last_revision_info, tags):
            from ....tag import DisabledTags, MemoryTags
            self.repository = repository
            self._last_revision_info = last_revision_info
            self._revision_history_cache = None
            if tags is not None:
                self.tags = MemoryTags(tags)
            else:
                self.tags = DisabledTags(self)
            self._partial_revision_history_cache = []
            self._last_revision_info_cache = None
            self._revision_id_to_revno_cache = None
            self._partial_revision_id_to_revno_cache = {}
            self._partial_revision_history_cache = []

        def lock_read(self):
            self.repository.lock_read()
            return LogicalLockResult(self.unlock)

        def lock_write(self, token=None):
            self.repository.lock_write()
            return BranchWriteLockResult(self.unlock, None)

        def unlock(self):
            self.repository.unlock()

        def last_revision_info(self):
            return self._last_revision_info

        def _gen_revision_history(self):
            """Generate the revision history from last revision
            """
            last_revno, last_revision = self.last_revision_info()
            self._extend_partial_history()
            return list(reversed(self._partial_revision_history_cache))

        def get_rev_id(self, revno, history=None):
            """Find the revision id of the specified revno."""
            with self.lock_read():
                if revno == 0:
                    return NULL_REVISION
                last_revno, last_revid = self.last_revision_info()
                if revno == last_revno:
                    return last_revid
                if last_revno is None:
                    self._extend_partial_history()
                    return self._partial_revision_history_cache[
                            len(self._partial_revision_history_cache) - revno]
                else:
                    if revno <= 0 or revno > last_revno:
                        raise NoSuchRevision(self, revno)
                    distance_from_last = last_revno - revno
                    if len(self._partial_revision_history_cache) <= distance_from_last:
                        self._extend_partial_history(distance_from_last)
                    return self._partial_revision_history_cache[distance_from_last]


class UpstreamBranchSource(UpstreamSource):
    """Upstream source that uses the upstream branch.

    :ivar upstream_branch: Branch with upstream sources
    :ivar upstream_version_map: Map from version strings to revspecs
    """

    def __init__(self, upstream_branch, upstream_revision_map=None,
                 config=None, actual_branch=None, create_dist=None,
                 other_repository=None, snapshot=True):
        self.upstream_branch = upstream_branch
        self._actual_branch = actual_branch or upstream_branch
        self.create_dist = create_dist
        self.config = config
        self.snapshot = snapshot
        self.other_repository = other_repository
        self.upstream_revision_map = {}
        if upstream_revision_map is not None:
            self.upstream_revision_map.update(upstream_revision_map.items())

    @classmethod
    def from_branch(cls, upstream_branch, upstream_revision_map=None,
                    config=None, local_dir=None, create_dist=None,
                    snapshot=True):
        """Create a new upstream branch source from a branch.

        This will optionally fetch into a local directory.
        """
        actual_branch = upstream_branch
        if local_dir is not None and not getattr(
                upstream_branch.repository, 'supports_random_access', True):
            local_repository = local_dir.find_repository()
            try:
                (last_revno,
                 last_revision) = upstream_branch.last_revision_info()
            except UnsupportedOperation:
                last_revno = None
                last_revision = upstream_branch.last_revision()
            local_repository.fetch(
                upstream_branch.repository, revision_id=last_revision)
            upstream_branch = MemoryBranch(
                local_repository, (last_revno, last_revision),
                upstream_branch.tags.get_tag_dict())
        return cls(
            upstream_branch=upstream_branch,
            upstream_revision_map=upstream_revision_map, config=config,
            actual_branch=actual_branch, create_dist=create_dist,
            snapshot=snapshot)

    def version_as_revision(self, package, version, tarballs=None):
        if version in self.upstream_revision_map:
            revspec = self.upstream_revision_map[version]
        else:
            revspec = get_export_upstream_revision(
                self.config, version=version)
        if revspec is not None:
            try:
                return RevisionSpec.from_string(
                    revspec).as_revision_id(self.upstream_branch)
            except (InvalidRevisionSpec, NoSuchTag):
                raise PackageVersionNotPresent(package, version, self)
        else:
            for revspec in guess_upstream_revspec(package, version):
                note(gettext('No upstream upstream-revision format '
                             'specified, trying %s') % revspec)
                try:
                    return RevisionSpec.from_string(
                        revspec).as_revision_id(self.upstream_branch)
                except (InvalidRevisionSpec, NoSuchTag):
                    pass
            else:
                raise PackageVersionNotPresent(package, version, self)
        raise PackageVersionNotPresent(package, version, self)

    def version_as_revisions(self, package, version, tarballs=None):
        # FIXME: Support multiple upstream locations if there are multiple
        # components
        if tarballs is not None and tarballs.keys() != [None]:
            raise MultipleUpstreamTarballsNotSupported()
        return {None: self.version_as_revision(package, version, tarballs)}

    def has_version(self, package, version, tarballs=None):
        try:
            self.version_as_revision(package, version, tarballs)
        except PackageVersionNotPresent:
            return False
        else:
            return True

    def get_latest_version(self, package, current_version):
        if self.snapshot:
            revid = self.upstream_branch.last_revision()
            version = self.get_version(package, current_version, revid)
            self.upstream_revision_map[version] = 'revid:%s' % revid.decode('utf-8')
            return version
        else:
            versions = list(self.get_recent_versions(package, current_version))
            if not versions:
                return None
            return versions[-1]

    def get_recent_versions(self, package, since_version=None):
        versions = []
        tags = self.upstream_branch.tags.get_tag_dict()
        with self.upstream_branch.repository.lock_read():
            graph = self.upstream_branch.repository.get_graph()
            if since_version is not None:
                since_revision = self.version_as_revision(
                    package, since_version)
            else:
                since_revision = None
            for tag, revision in tags.items():
                version = upstream_tag_to_version(tag, package)
                if version is None:
                    continue
                if since_version is not None and version <= since_version:
                    continue
                if since_revision and not graph.is_ancestor(
                        since_revision, revision):
                    continue
                versions.append(version)
        return sorted(versions)

    def get_version(self, package, current_version, revision):
        with self.upstream_branch.lock_read():
            return upstream_branch_version(
                self.upstream_branch, revision, package, current_version)

    def fetch_tarballs(self, package, version, target_dir, components=None,
                       revisions=None):
        if components is not None and components != [None]:
            # Multiple components are not supported
            raise PackageVersionNotPresent(package, version, self)
        note("Looking for upstream %s/%s in upstream branch %r.",
             package, version, self.upstream_branch)
        with self.upstream_branch.lock_read():
            if revisions is not None:
                revid = revisions[None]
            else:
                revid = self.version_as_revision(package, version)
                if revid is None:
                    raise PackageVersionNotPresent(package, version, self)
            if self.other_repository is not None:
                try:
                    rev_tree = self.other_repository.revision_tree(revid)
                except NoSuchRevision:
                    rev_tree = None
            else:
                rev_tree = None
            if rev_tree is None:
                rev_tree = self.upstream_branch.repository.revision_tree(revid)
            if self.create_dist is not None:
                with tempfile.TemporaryDirectory() as td:
                    fn = self.create_dist(rev_tree, package, version, td)
                    if fn:
                        nfn = new_tarball_name(package, version, fn)
                        repack_tarball(os.path.join(td, fn), nfn, target_dir)
                        return [os.path.join(target_dir, nfn)]
            tarball_base = "%s-%s" % (package, version)
            target_filename = self._tarball_path(
                package, version, None, target_dir)
            try:
                export(rev_tree, target_filename, 'tgz', tarball_base)
            except UnsupportedOperation as e:
                note('Not exporting revision from upstream branch: %s', e)
                raise PackageVersionNotPresent(package, version, self)
            else:
                note("Exporting upstream branch revision %s to create "
                     "the tarball", revid)
        return [target_filename]

    def __repr__(self):
        return "<%s for %r>" % (
            self.__class__.__name__, self._actual_branch.base)


class LazyUpstreamBranchSource(UpstreamBranchSource):
    """Upstream branch source that defers loading the branch until it is used.
    """

    def __init__(self, upstream_branch_url, upstream_revision_map=None,
                 config=None, create_dist=None, other_repository=None,
                 snapshot=True):
        self.upstream_branch_url = upstream_branch_url
        self.snapshot = snapshot
        self._upstream_branch = None
        self.config = config
        self.create_dist = create_dist
        self.other_repository = other_repository
        if upstream_revision_map is None:
            self.upstream_revision_map = {}
        else:
            self.upstream_revision_map = upstream_revision_map

    @property
    def upstream_branch(self):
        if self._upstream_branch is None:
            if callable(self.upstream_branch_url):
                self.upstream_branch_url = self.upstream_branch_url()
            self._upstream_branch = Branch.open(self.upstream_branch_url)
        return self._upstream_branch

    def __repr__(self):
        return "<%s for %r>" % (
            self.__class__.__name__, self.upstream_branch_url)


class DistCommandFailed(BzrError):

    _fmt = "Dist command failed to produce a tarball: %(error)s"

    def __init__(self, error):
        super(DistCommandFailed, self).__init__(error=error)


def run_dist_command(
        rev_tree: Tree, package: str, version: Version, target_dir: str,
        dist_command: str) -> bool:
    with tempfile.TemporaryDirectory() as td:
        package_dir = os.path.join(td, package)
        export(rev_tree, package_dir, 'dir')
        existing_files = os.listdir(package_dir)
        env = dict(os.environ.items())
        env['PACKAGE'] = package
        env['VERSION'] = version
        note('Running dist command: %s', dist_command)
        try:
            subprocess.check_call(
                dist_command, env=env, cwd=package_dir, shell=True)
        except subprocess.CalledProcessError as e:
            if e.returncode == 2:
                return None
            raise DistCommandFailed(str(e))
        new_files = os.listdir(package_dir)
        diff_files = set(new_files) - set(existing_files)
        diff = [n for n in diff_files if get_filetype(n) is not None]
        if len(diff) == 1:
            note('Found tarball %s in package directory.', diff[0])
            os.rename(
                os.path.join(package_dir, diff[0]),
                os.path.join(target_dir, diff[0]))
            return diff[0]
        if 'dist' in diff_files:
            for entry in os.scandir(os.path.join(package_dir, 'dist')):
                if get_filetype(entry.name) is not None:
                    note('Found tarball %s in dist directory.', entry.name)
                    os.rename(entry.path, os.path.join(target_dir, entry.name))
                    return entry.name
            note('No tarballs found in dist directory.')
        diff = set(os.listdir(td)) - set([package])
        if len(diff) == 1:
            fn = diff.pop()
            note('Found tarball %s in parent directory.', fn)
            os.rename(
                os.path.join(td, fn),
                os.path.join(target_dir, fn))
            return fn
    raise DistCommandFailed('no tarball created')



#    upstream.py -- Providers of upstream source
#    Copyright (C) 2009 Canonical Ltd.
#    Copyright (C) 2009 Jelmer Vernooij <jelmer@debian.org>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import re
import shutil
import subprocess
import sys
import tarfile
import tempfile

from debian.changelog import Version

from ....errors import BzrError, DependencyNotPresent, NoSuchFile
from .... import osutils
from ....export import export
from ....trace import (
    note,
    warning,
    )

from ..repack_tarball import (
    get_filetype,
    repack_tarball,
    )
from ..util import (
    component_from_orig_tarball,
    tarball_name,
    )


class PackageVersionNotPresent(BzrError):
    _fmt = "%(package)s %(version)s was not found in %(upstream)s."

    def __init__(self, package, version, upstream):
        BzrError.__init__(self, package=package, version=version,
                          upstream=upstream)


class MissingUpstreamTarball(BzrError):
    _fmt = ("Unable to find the needed upstream tarball for package "
            "%(package)s, version %(version)s.")

    def __init__(self, package, version):
        BzrError.__init__(self, package=package, version=version)


def new_tarball_name(package, version, old_name):
    """Determine new tarball name based on the package name and the old name.
    """
    old_format = get_filetype(old_name)
    if old_format in ("gz", "bz2", "xz"):
        return tarball_name(package, version, None, old_format)
    # Unknown target format, repack to .tar.gz
    return tarball_name(package, version, None, "gz")


class UpstreamSource(object):
    """A source for upstream versions (uscan, debian/rules, etc)."""

    def get_latest_version(self, package, current_version):
        """Check what the latest upstream version is.

        :param package: Name of the package
        :param version: The current upstream version of the package.
        :return: The version string of the latest available upstream version.
        """
        raise NotImplementedError(self.get_latest_version)

    def get_recent_versions(self, package, since_version=None):
        """Retrieve recent version strings.

        :param package: Name of the package
        :param version: Last upstream version since which to retrieve versions
        :return: Iterator over version strings
        """
        raise NotImplementedError(self.get_recent_versions)

    def version_as_revisions(self, package, version, tarballs=None):
        """Lookup the revision ids for a particular version.

        :param package: Package name
        :param version: Version string
        :raise PackageVersionNotPresent: When the specified version was not
            found
        :return: dictionary mapping component names to revision ids
        """
        raise NotImplementedError(self.version_as_revisions)

    def has_version(self, package, version, tarballs=None):
        """Check whether this upstream source contains a particular package.

        :param package: Package name
        :param version: Version string
        :param tarballs: Tarballs list
        """
        raise NotImplementedError(self.has_version)

    def fetch_tarballs(self, package, version, target_dir, components=None):
        """Fetch the source tarball for a particular version.

        :param package: Name of the package
        :param version: Version string of the version to fetch
        :param target_dir: Directory in which to store the tarball
        :param components: List of component names to fetch; may be None,
            in which case the backend will have to find out.
        :return: Paths of the fetched tarballs
        """
        raise NotImplementedError(self.fetch_tarballs)

    def _tarball_path(self, package, version, component, target_dir,
                      format=None):
        return os.path.join(
            target_dir,
            tarball_name(package, version, component, format=format))


class AptSource(UpstreamSource):
    """Upstream source that uses apt-source."""

    def fetch_tarballs(self, package, upstream_version, target_dir,
                       _apt_pkg=None, components=None):
        if _apt_pkg is None:
            try:
                import apt_pkg
            except ImportError as e:
                raise DependencyNotPresent('apt_pkg', e)
        else:
            apt_pkg = _apt_pkg
        apt_pkg.init()

        # Handle the case where the apt.sources file contains no source
        # URIs (LP:375897)
        try:
            sources = apt_pkg.SourceRecords()
        except SystemError:
            raise PackageVersionNotPresent(package, upstream_version, self)

        sources.restart()
        note("Using apt to look for the upstream tarball.")
        while sources.lookup(package):
            filenames = []
            for (checksum, size, filename, filekind) in sources.files:
                if filekind != "tar":
                    continue
                filename = os.path.basename(filename)
                if filename.startswith(
                        "%s_%s.orig" % (package, upstream_version)):
                    filenames.append(filename)
            if filenames:
                if self._run_apt_source(package, sources.version, target_dir):
                    return [os.path.join(target_dir, filename)
                            for filename in filenames]
                else:
                    note("apt found %s/%s but could not download.",
                         package, sources.version)
        note("apt could not find %s/%s.", package, upstream_version)
        raise PackageVersionNotPresent(package, upstream_version, self)

    def _get_command(self, package, version_str):
        return 'apt-get source -y --only-source --tar-only %s=%s' % \
            (package, version_str)

    def _run_apt_source(self, package, version_str, target_dir):
        command = self._get_command(package, version_str)
        proc = subprocess.Popen(command, shell=True, cwd=target_dir)
        proc.wait()
        if proc.returncode != 0:
            return False
        return True


class SelfSplitSource(UpstreamSource):

    def __init__(self, tree):
        self.tree = tree

    def _split(self, package, upstream_version, target_filename):
        with tempfile.TemporaryDirectory(
                prefix="builddeb-get-orig-source-") as tmpdir:
            export_dir = os.path.join(
                tmpdir, "%s-%s" % (package, upstream_version))
            export(self.tree, export_dir, format="dir")
            shutil.rmtree(os.path.join(export_dir, "debian"))
            with tarfile.open(target_filename, "w:gz") as tar:
                tar.add(export_dir, "%s-%s" % (package, upstream_version))

    def fetch_tarballs(self, package, version, target_dir, components=None):
        note("Using the current branch without the 'debian' directory "
             "to create the tarball")
        tarball_path = self._tarball_path(package, version, None, target_dir)
        self._split(package, version, tarball_path)
        return [tarball_path]


class StackedUpstreamSource(UpstreamSource):
    """An upstream source that checks a list of other upstream sources.

    The first source that can provide a tarball, wins.
    """

    def __init__(self, sources):
        self._sources = sources

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self._sources)

    def fetch_tarballs(self, package, version, target_dir, components=None):
        for source in self._sources:
            try:
                paths = source.fetch_tarballs(
                    package, version, target_dir, components)
            except PackageVersionNotPresent:
                pass
            except DependencyNotPresent as e:
                warning('not checking %r due to missing dependency: %s',
                        source, e)
            else:
                return paths
        raise PackageVersionNotPresent(package, version, self)

    def get_latest_version(self, package, version):
        for source in self._sources:
            try:
                new_version = source.get_latest_version(package, version)
                if new_version is not None:
                    return new_version
            except NotImplementedError:
                pass
        return None

    def get_recent_versions(self, package, since_version=None):
        versions = set()
        for source in self._sources:
            for version in source.get_recent_versions(package, since_version):
                versions.add(version)
        return list(sorted(versions))

    def version_as_revisions(self, package, version, tarballs=None):
        for source in self._sources:
            try:
                return source.version_as_revisions(package, version, tarballs)
            except PackageVersionNotPresent:
                pass
            except DependencyNotPresent as e:
                warning('not checking %r due to missing dependency: %s',
                        source, e)
        raise PackageVersionNotPresent(package, version, self)

    def has_version(self, package, version, tarballs=None):
        for source in self._sources:
            if source.has_version(package, version, tarballs):
                return True
        return False


def gather_orig_files(package, version, path):
    """Grab the orig files for a particular package.

    :param package: package name
    :param version: package upstream version string
    :return: List of orig tarfile paths, or None if none were found
    """
    prefix = ("%s_%s.orig" % (package, version))
    ret = []
    path = os.path.abspath(path)
    if not os.path.isdir(path):
        return None
    if isinstance(path, bytes):
        prefix = prefix.encode(osutils._fs_enc)
    for filename in os.listdir(path):
        if filename.endswith('.asc'):
            continue
        if filename.startswith(prefix):
            ret.append(os.path.join(path, filename))
    if ret:
        return ret
    return None


class UpstreamProvider(object):
    """An upstream provider can provide the upstream source for a package.

    Different implementations can provide it in different ways, for
    instance using pristine-tar, or using apt.
    """

    def __init__(self, package, version, store_dir, sources):
        """Create an UpstreamProvider.

        :param package: the name of the source package that is being built.
        :param version: the Version of the package that is being built.
        :param store_dir: A directory to cache the tarballs.
        """
        self.package = package
        self.version = version
        self.store_dir = os.path.abspath(store_dir)
        self.source = StackedUpstreamSource(sources)

    def provide(self, target_dir):
        """Provide the upstream tarball(s) any way possible.

        Call this to place the correctly named tarball in to target_dir,
        through means possible.

        If the tarball is already there then do nothing.
        If it is in self.store_dir then copy it over.
        Else retrive it and cache it in self.store_dir, then copy it over:
           - If pristine-tar metadata is available then that will be used.
           - Else if apt knows about a source of that version that will be
             retrieved.
           - Else if uscan knows about that version it will be downloaded
             and repacked as needed.
           - Else a call will be made to debian/rules to try and retrieve
             the tarball.

        If the tarball can't be found at all then MissingUpstreamTarball
        will be raised.

        :param target_dir: The directory to place the tarball in.
        :return: The path to the tarball.
        """
        note("Looking for a way to retrieve the upstream tarball")
        in_target = self.already_exists_in_target(target_dir)
        if in_target is not None:
            note("Upstream tarball already exists in build directory, "
                 "using that")
            return [
                (p, component_from_orig_tarball(p, self.package, self.version))
                for p in in_target]
        if self.already_exists_in_store() is None:
            if not os.path.exists(self.store_dir):
                os.makedirs(self.store_dir)
            try:
                paths = self.source.fetch_tarballs(
                    self.package, self.version, self.store_dir)
            except PackageVersionNotPresent:
                raise MissingUpstreamTarball(self.package, self.version)
            assert isinstance(paths, list)
        else:
            note("Using the upstream tarball that is present in %s" %
                 self.store_dir)
        paths = self.provide_from_store_dir(target_dir)
        assert paths is not None
        return [(p, component_from_orig_tarball(p, self.package, self.version))
                for p in paths]

    def already_exists_in_target(self, target_dir):
        return gather_orig_files(self.package, self.version, target_dir)

    def already_exists_in_store(self):
        return gather_orig_files(self.package, self.version, self.store_dir)

    def provide_from_store_dir(self, target_dir):
        paths = self.already_exists_in_store()
        if paths is None:
            return None
        for path in paths:
            repack_tarball(
                path, os.path.basename(path), target_dir=target_dir)
        return paths


def extract_tarball_version(path, packagename):
    """Extract a version from a tarball path.

    :param path: Path to the tarball (e.g. "/tmp/bzr-builddeb-2.7.3.tar.gz")
    :param packagename: Name of the package (e.g. "bzr-builddeb")
    """
    basename = os.path.basename(path)
    for extension in [".tar.gz", ".tgz", ".tar.bz2", ".tar.lzma", ".tar.xz",
                      ".zip"]:
        if basename.endswith(extension):
            basename = basename[:-len(extension)]
            break
    else:
        # Unknown extension
        return None
    # Debian style tarball
    m = re.match(packagename+"_(.*).orig", basename)
    if m:
        return m.group(1)
    # Traditional, PACKAGE-VERSION.tar.gz
    m = re.match(packagename+"-(.*)", basename)
    if m:
        return m.group(1)
    return None


class TarfileSource(UpstreamSource):
    """Source that uses a single local tarball."""

    def __init__(self, path, version=None):
        self.path = path
        if version is not None:
            self.version = str(version)
        else:
            self.version = None

    def fetch_tarballs(self, package, version, target_dir, components=None):
        if version != self.version:
            raise PackageVersionNotPresent(package, version, self)
        dest_name = new_tarball_name(package, version, self.path)
        repack_tarball(self.path, dest_name, target_dir=target_dir)
        return [os.path.join(target_dir, dest_name)]

    def get_recent_versions(self, package, since_version=None):
        latest_version = self.get_latest_version(package, since_version)
        return [latest_version]

    def get_latest_version(self, package, version):
        if self.version is not None:
            return self.version
        self.version = extract_tarball_version(self.path, package)
        return self.version


class LaunchpadReleaseFileSource(UpstreamSource):
    """Source that retrieves release files from Launchpad."""

    @classmethod
    def from_package(cls, distribution_name, distroseries_name, package):
        """Create a LaunchpadReleaseFileSource from a distribution package.

        :param distribution_name: Name of the distribution (e.g. "Ubuntu")
        :param distroseries_name: Name of the distribution series
            (e.g. "oneiric")
        :param package: Package name
        :return: A `LaunchpadReleaseFileSource`
        """
        from ...launchpad import (
            get_upstream_projectseries_for_package,
            )
        project_series = get_upstream_projectseries_for_package(
            package, distribution_name, distroseries_name)
        if project_series is None:
            return None
        return cls(project_series=project_series)

    def __init__(self, project=None, project_series=None):
        if project_series is None:
            self.project_series = project.development_focus
        else:
            self.project_series = project_series
        if project is None:
            self.project = project_series.project
        else:
            self.project = project

    def fetch_tarballs(self, package, version, target_dir, components=None):
        note("Retrieving tarball for %s from Launchpad.", package)
        release = self.project.getRelease(version=version)
        if release is None:
            raise PackageVersionNotPresent(package, version, self)
        release_files = []
        for f in release.files:
            if f.file_type == "Code Release Tarball":
                release_files.append(f.file)
        if len(release_files) == 0:
            warning("Release %s for package %s found on Launchpad but no "
                    "associated tarballs.", version, package)
            raise PackageVersionNotPresent(package, version, self)
        elif len(release_files) > 1:
            warning("More than one release file for release %s of package %s"
                    "found on Launchpad. Using the first.", version, package)
        hosted_file = release_files[0]
        with tempfile.TemporaryDirectory(
                prefix="builddeb-launchpad-source-") as tmpdir:
            with hosted_file.open() as inf:
                note("Downloading upstream tarball %s from Launchpad",
                     inf.filename)
                filename = inf.filename.encode(osutils._fs_enc)
                filename = filename.replace("/", "")
                tmppath = os.path.join(tmpdir, filename)
                with open(tmppath, 'wb') as outf:
                    outf.write(inf.read())
            dest_name = new_tarball_name(package, version, filename)
            repack_tarball(tmppath, dest_name, target_dir=target_dir)
            return os.path.join(target_dir, dest_name)

    def _all_versions(self):
        for release in self.project_series.releases:
            yield (release.date_released, release.version)

    def get_recent_versions(self, package, since_version=None):
        versions = []
        for version in self._all_versions():
            if since_version is None or since_version < version:
                versions.append(version)
        return sorted(versions)

    def get_latest_version(self, package, version):
        versions = list(self._all_versions())
        versions.sort()
        return versions[-1][1]


class DirectoryScanSource(UpstreamSource):
    """Source that scans a local directory for sources."""

    def __init__(self, path):
        self.path = os.path.abspath(path)

    def fetch_tarballs(self, package, version, target_dir, components=None):
        v = Version(version)
        v.epoch = None
        prefix = '%s_%s.orig' % (package, v)
        ret = []
        for entry in os.scandir(self.path):
            if entry.name.startswith(prefix):
                component = (entry.name[len(prefix):].split('.')[0] or None)
                shutil.copy(entry.path, target_dir)
                if components is None or component in components:
                    ret.append(os.path.join(target_dir, entry.name))
        if ret:
            return ret
        raise PackageVersionNotPresent(package, version, self)

    def get_recent_versions(self, package, since_version=None):
        return []

    def get_latest_version(self, package, version):
        return None
